? test/regression/NativeLibTest.java
Index: kaffe/kaffevm/classPool.c
===================================================================
RCS file: /cvs/kaffe/kaffe/kaffe/kaffevm/classPool.c,v
retrieving revision 1.31
diff -u -p -r1.31 classPool.c
--- kaffe/kaffevm/classPool.c	19 Dec 2004 06:25:07 -0000	1.31
+++ kaffe/kaffevm/classPool.c	19 Dec 2004 14:26:26 -0000
@@ -242,6 +242,7 @@ destroyClassLoader(Collector *c UNUSED, 
         int rmoved;
 
         rmoved = removeClassEntries(loader);
+	unloadNativeLibraries(loader);
 
 	if (Kaffe_JavaVMArgs.enableVerboseGC > 0) {
 		dprintf("<GC: destroying classloader "
Index: kaffe/kaffevm/external.c
===================================================================
RCS file: /cvs/kaffe/kaffe/kaffe/kaffevm/external.c,v
retrieving revision 1.71
diff -u -p -r1.71 external.c
--- kaffe/kaffevm/external.c	19 Dec 2004 06:25:08 -0000	1.71
+++ kaffe/kaffevm/external.c	19 Dec 2004 14:26:26 -0000
@@ -40,6 +40,7 @@
 #include "jthread.h"
 #include "jsignal.h"
 #include "stats.h"
+#include "locks.h"
 #if defined(KAFFE_FEEDBACK)
 #include "feedback.h"
 #endif
@@ -68,46 +69,17 @@
 static struct _libHandle {
 	LIBRARYHANDLE	desc;
 	char*		name;
-	int		ref;
+	Hjava_lang_ClassLoader*		loader;
 } libHandle[MAXLIBS];
 
-static inline void *
-findLibraryFunction(const char *name) {
-  int i = 0;
-  void *ptr = NULL;
-  
-  while (!ptr && libHandle[i].ref && i < MAXLIBS) {
-    ptr = KaffeLib_GetSymbol(libHandle[i].desc, name);
-
-DBG(NATIVELIB,
-    if (ptr == NULL) {
-	dprintf("Couldn't find %s in library handle %d == %s.\nError message is %s.\n",
-		name,
-		i,
-		lt_dlgetinfo(libHandle[i].desc) == NULL ? "unknown" : lt_dlgetinfo(libHandle[i].desc)->name,
-		lt_dlerror());
-    }
-    else {
-	dprintf("Found %s in library handle %d == %s.\n",
-		name,
-        	i,
-        	lt_dlgetinfo(libHandle[i].desc) == NULL ? "unknown" : lt_dlgetinfo(libHandle[i].desc)->name);
-    }
-);
-
-    ++i;
-  }
-
-  return ptr;
-}
-
-static char *libraryPath = "";
+static iStaticLock	libraryLock = KAFFE_STATIC_LOCK_INITIALIZER; /* mutex on all intern operations */
+static const char *libraryPath = "";
 
 extern JavaVM Kaffe_JavaVM;
 
 #ifdef ENABLE_BINRELOC
 static
-const char *discoverLibraryHome()
+const char *discoverLibraryHome(void)
 {
   static char libraryHome[MAXPATHLEN];
   char *entryPoint;
@@ -206,7 +178,7 @@ initNative(void)
 
 	       	DBG(INIT, dprintf("trying to load %s\n", lib); );
 
-		if (loadNativeLibrary(lib, NULL, 0) >= 0) {
+		if (loadNativeLibrary(lib, NULL, NULL, 0) >= 0) {
 			DBG(INIT, dprintf("initNative() done\n"); );
 			return;
 		}
@@ -218,26 +190,20 @@ initNative(void)
 	EXIT(1);
 }
 
-int
-loadNativeLibrary(char* lib, char *errbuf, size_t errsiz)
-{
-	int retval;
-
-	retval = loadNativeLibrary2(lib, 1, errbuf, errsiz);
-	return( retval );
-}
-
 /*
  * Link in a native library. If successful, returns an index >= 0 that
  * can be passed to unloadNativeLibrary(). Otherwise, returns -1 and
  * fills errbuf (if not NULL) with the error message. Assumes synchronization.
  */
 int
-loadNativeLibrary2(char* path, int default_refs, char *errbuf, size_t errsiz)
+loadNativeLibrary(char* path, struct Hjava_lang_ClassLoader* loader, char *errbuf, size_t errsiz)
 {
 	struct _libHandle *lib;
 	int libIndex;
 	void *func;
+	int iLockRoot;
+
+	lockStaticMutex(&libraryLock);
 
 	/* Find a library handle.  If we find the library has already
 	 * been loaded, don't bother to get it again, just increase the
@@ -248,15 +214,24 @@ loadNativeLibrary2(char* path, int defau
 		if (lib->desc == 0) {
 			goto open;
 		}
-		if (strcmp(lib->name, path) == 0) {
-			lib->ref++;
+		if (strcmp(lib->name, path) != 0)
+			continue;
+
+		if (lib->loader != loader) {
+			if (errbuf != NULL) {
+				strncpy(errbuf, "Already loaded\n", errsiz);
+			}
+			unlockStaticMutex(&libraryLock);
+			return -1;
+		}
+
 DBG(NATIVELIB,
 			dprintf("Native lib %s\n"
-			    "\tLOAD desc=%p index=%d ++ref=%d\n",
-			    lib->name, lib->desc, libIndex, lib->ref);
+			    "\tLOAD desc=%p index=%d loader=%p\n",
+			    lib->name, lib->desc, libIndex, lib->loader);
     );
-			return libIndex;
-		}
+		unlockStaticMutex(&libraryLock); 
+		return libIndex;
 	}
 	if (errbuf != 0) {
 		assert(errsiz > 0);
@@ -264,6 +239,7 @@ DBG(NATIVELIB,
 		strncpy(errbuf, "Too many open libraries", errsiz);
 		errbuf[errsiz - 1] = '\0';
 	}
+	unlockStaticMutex(&libraryLock);
 	return -1;
 
 	/* Open the library */
@@ -283,10 +259,9 @@ DBG(NATIVELIB,
 /* if we tested for existence here, libltdl wouldn't be able to look
    for system-dependent library names */
 
-	blockAsyncSignals();
 	{
                 lib->desc = KaffeLib_Load(path);
-                if (lib->desc == 0)	
+                if (lib->desc == NULL)	
 			{
 				const char *err = KaffeLib_GetError();
 
@@ -322,25 +297,26 @@ DBG(NATIVELIB,
 					if (errbuf != 0)
 						strncpy(errbuf, err, errsiz);
 				}
+			
+				unlockStaticMutex(&libraryLock);
+				return -1;
 			}
 		}
-	unblockAsyncSignals();
-
-	if (lib->desc == 0) {
-		return -1;
-	}
 
         lib->name = gc_malloc(strlen(path)+1, KGC_ALLOC_NATIVELIB);
         strcpy (lib->name, path);
 
-	lib->ref = default_refs;
+	lib->loader = loader;
 	addToCounter(&ltmem, "vmmem-libltdl", 1, GCSIZEOF(lib->name));
 
+	unlockStaticMutex(&libraryLock);
+
 DBG(NATIVELIB,
 	dprintf("Native lib %s\n"
-	    "\tLOAD desc=%p index=%d ++ref=%d\n",
-	    lib->name, lib->desc, libIndex, lib->ref);
+	    "\tLOAD desc=%p index=%d loader=%p\n",
+	    lib->name, lib->desc, libIndex, lib->loader);
     );
+
 #if defined(KAFFE_FEEDBACK)
 	feedbackLibrary(path, true);
 #endif
@@ -361,28 +337,37 @@ DBG(NATIVELIB,
  * never be unloaded. So index should never equal zero here.
  */
 void
-unloadNativeLibrary(int libIndex)
+unloadNativeLibraries(struct Hjava_lang_ClassLoader* loader)
 {
-	struct _libHandle *lib;
+	int libIndex;
+	int iLockRoot;
+
+	lockStaticMutex(&libraryLock);
+
+	/* we should never ever unload libraries of the bootstrap loader */
+	assert (loader != NULL);
+
+	for (libIndex = 0; libIndex < MAXLIBS; libIndex++) {
+		struct _libHandle* lib = &libHandle[libIndex];
+
+		if (lib->desc == NULL)
+			continue;
 
-	assert(libIndex > 0 && libIndex < MAXLIBS);
-	lib = &libHandle[libIndex];
+		if (lib->loader != loader)
+			continue;
 
 DBG(NATIVELIB,
-	dprintf("Native lib %s\n"
-	    "\tUNLOAD desc=%p index=%d --ref=%d\n",
-	    lib->name, lib->desc, libIndex, lib->ref - 1);
+    dprintf("Native lib %s\n"
+	    "\tUNLOAD desc=%p index=%d loader=%p\n",
+	    lib->name, lib->desc, libIndex, lib->loader);
     );
 
-	assert(lib->desc != 0);
-	assert(lib->ref > 0);
-	if (--lib->ref == 0) {
-		blockAsyncSignals();
 		KaffeLib_Unload(lib->desc);
-		unblockAsyncSignals();
 		KFREE(lib->name);
-		lib->desc = 0;
+		lib->desc = NULL;
 	}
+
+	unlockStaticMutex(&libraryLock);
 }
 
 /*
@@ -393,10 +378,11 @@ loadNativeLibrarySym(const char* name)
 {
   int i = 0;
   void* func = NULL;;
+  int iLockRoot;
 
-  blockAsyncSignals();
+  lockStaticMutex(&libraryLock);
   
-  while (!func && libHandle[i].ref && i < MAXLIBS) {
+  while (!func && i < MAXLIBS && libHandle[i].desc!=NULL) {
     func = KaffeLib_GetSymbol(libHandle[i].desc, name);
     
 DBG(NATIVELIB,
@@ -418,7 +404,7 @@ DBG(NATIVELIB,
     ++i;
   }
 
-  unblockAsyncSignals();
+  unlockStaticMutex(&libraryLock);
 
   return func;
 }
Index: kaffe/kaffevm/gcFuncs.c
===================================================================
RCS file: /cvs/kaffe/kaffe/kaffe/kaffevm/gcFuncs.c,v
retrieving revision 1.59
diff -u -p -r1.59 gcFuncs.c
--- kaffe/kaffevm/gcFuncs.c	19 Dec 2004 06:25:08 -0000	1.59
+++ kaffe/kaffevm/gcFuncs.c	19 Dec 2004 14:26:26 -0000
@@ -376,6 +376,8 @@ DBG(GCPRECISE,
                 walkMethods(collector, gc_info, CLASS_METHODS(class), CLASS_NMETHODS(class));
         }
         KGC_markObject(collector, gc_info, class->loader);
+	KGC_markObject(collector, gc_info, class->signers);
+	KGC_markObject(collector, gc_info, class->protectionDomain);
 }
 
 /*****************************************************************************
Index: kaffe/kaffevm/thread.c
===================================================================
RCS file: /cvs/kaffe/kaffe/kaffe/kaffevm/thread.c,v
retrieving revision 1.85
diff -u -p -r1.85 thread.c
--- kaffe/kaffevm/thread.c	19 Dec 2004 06:25:08 -0000	1.85
+++ kaffe/kaffevm/thread.c	19 Dec 2004 14:26:26 -0000
@@ -249,12 +249,19 @@ attachFakedThreadInstance(const char* nm
         /*
 	 * set context class loader of primordial thread to app classloader
 	 * must not be done earlier, since getCurrentThread() won't work
-         * before the KTHREAD(createfirst) and the jthreadID assignment
+         * before the KTHREAD(createfirst) and the jthreadID assignment.
+	 *
+	 * If we're creating the Thread instance of the main thread, this
+	 * will trigger the initialization process of the java part of the
+	 * runtime. One pitfall during initialization is that java.lang.System
+	 * has to be initialized before kaffe.lang.AppClassLoader. Therefore,
+	 * we cannot call AppClassLoader.getSingleton() here, since that would
+	 * cause System and AppClassLoader to be initialized in the wrong order.
 	 */
-	  do_execute_java_class_method (&retval, "kaffe/lang/AppClassLoader",
-					NULL,
-					"getSingleton",
-					"()Ljava/lang/ClassLoader;");
+	do_execute_java_class_method (&retval, "java/lang/ClassLoader",
+				      NULL,
+				      "getSystemClassLoader",
+				      "()Ljava/lang/ClassLoader;");
         unhand(tid)->contextClassLoader = (struct Hjava_lang_ClassLoader *) retval.l;
 
 	/* Attach thread to threadGroup */
Index: kaffe/kaffevm/classMethod.h
===================================================================
RCS file: /cvs/kaffe/kaffe/kaffe/kaffevm/classMethod.h,v
retrieving revision 1.70
diff -u -p -r1.70 classMethod.h
--- kaffe/kaffevm/classMethod.h	5 Dec 2004 22:43:28 -0000	1.70
+++ kaffe/kaffevm/classMethod.h	19 Dec 2004 14:26:27 -0000
@@ -155,6 +155,7 @@ struct Hjava_lang_Class {
 	short			nr_inner_classes;
 	struct _innerClass*	inner_classes;
 
+	/* misc other stuff */
 	void*			gcjPeer;	/* only needed if GCJ_SUPPORT */
 #ifdef KAFFE_VMDEBUG
 	int			live_count;
@@ -166,6 +167,9 @@ struct Hjava_lang_Class {
 
 	/** The array of 'signer' objects; usually Certificates. */
 	HArrayOfObject*		signers;
+
+	/** The protection domain */
+	struct Hjava_lang_Object*	protectionDomain;
 };
 
 #ifndef __DEFINED_CLASS
Index: kaffe/kaffevm/external.h
===================================================================
RCS file: /cvs/kaffe/kaffe/kaffe/kaffevm/external.h,v
retrieving revision 1.9
diff -u -p -r1.9 external.h
--- kaffe/kaffevm/external.h	8 Dec 2004 07:22:07 -0000	1.9
+++ kaffe/kaffevm/external.h	19 Dec 2004 14:26:27 -0000
@@ -28,9 +28,8 @@ struct _methods;
 struct _errorInfo;
 
 void	initNative(void);
-int	loadNativeLibrary(char*, char*, size_t);
-int	loadNativeLibrary2(char*, int, char*, size_t);
-void	unloadNativeLibrary(int);
+int	loadNativeLibrary(char*, struct Hjava_lang_ClassLoader*, char*, size_t);
+void	unloadNativeLibraries(struct Hjava_lang_ClassLoader*);
 void*	loadNativeLibrarySym(const char*);
 nativecode*	native(struct _methods*, struct _errorInfo*);
 void	addNativeFunc(char*, void*);
Index: libraries/javalib/Makefile.am
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/Makefile.am,v
retrieving revision 1.275
diff -u -p -r1.275 Makefile.am
--- libraries/javalib/Makefile.am	13 Dec 2004 00:38:43 -0000	1.275
+++ libraries/javalib/Makefile.am	19 Dec 2004 14:26:28 -0000
@@ -1388,6 +1388,7 @@ gnu_java_text_SRCS = \
 	gnu/java/text/StringFormatBuffer.java \
 	gnu/java/text/WordBreakIterator.java
 gnu_java_util_SRCS = \
+	gnu/java/util/DoubleEnumeration.java \
 	gnu/java/util/EmptyEnumeration.java
 gnu_java_util_prefs_SRCS = \
 	gnu/java/util/prefs/FileBasedFactory.java \
@@ -2180,7 +2181,6 @@ java_lang_SRCS = \
 	java/lang/LinkageError.java \
 	java/lang/Long.java \
 	java/lang/Math.java \
-	java/lang/NativeLibrary.java \
 	java/lang/NegativeArraySizeException.java \
 	java/lang/NoClassDefFoundError.java \
 	java/lang/NoSuchFieldError.java \
@@ -2220,6 +2220,7 @@ java_lang_SRCS = \
 	java/lang/UnsupportedOperationException.java \
 	java/lang/VerifyError.java \
 	java/lang/VirtualMachineError.java \
+	java/lang/VMClassLoader.java \
 	java/lang/VMCompiler.java \
 	java/lang/VMObject.java \
 	java/lang/VMSecurityManager.java \
Index: libraries/javalib/Makefile.in
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/Makefile.in,v
retrieving revision 1.356
diff -u -p -r1.356 Makefile.in
--- libraries/javalib/Makefile.in	13 Dec 2004 00:38:43 -0000	1.356
+++ libraries/javalib/Makefile.in	19 Dec 2004 14:26:30 -0000
@@ -1826,6 +1826,7 @@ gnu_java_text_SRCS = \
 	gnu/java/text/WordBreakIterator.java
 
 gnu_java_util_SRCS = \
+	gnu/java/util/DoubleEnumeration.java \
 	gnu/java/util/EmptyEnumeration.java
 
 gnu_java_util_prefs_SRCS = \
@@ -2653,7 +2654,6 @@ java_lang_SRCS = \
 	java/lang/LinkageError.java \
 	java/lang/Long.java \
 	java/lang/Math.java \
-	java/lang/NativeLibrary.java \
 	java/lang/NegativeArraySizeException.java \
 	java/lang/NoClassDefFoundError.java \
 	java/lang/NoSuchFieldError.java \
@@ -2693,6 +2693,7 @@ java_lang_SRCS = \
 	java/lang/UnsupportedOperationException.java \
 	java/lang/VerifyError.java \
 	java/lang/VirtualMachineError.java \
+	java/lang/VMClassLoader.java \
 	java/lang/VMCompiler.java \
 	java/lang/VMObject.java \
 	java/lang/VMSecurityManager.java \
Index: libraries/javalib/all.files
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/all.files,v
retrieving revision 1.57
diff -u -p -r1.57 all.files
--- libraries/javalib/all.files	8 Dec 2004 20:38:42 -0000	1.57
+++ libraries/javalib/all.files	19 Dec 2004 14:26:32 -0000
@@ -1719,7 +1719,6 @@ java/lang/InterruptedException.java
 java/lang/LinkageError.java
 java/lang/Long.java
 java/lang/Math.java
-java/lang/NativeLibrary.java
 java/lang/NegativeArraySizeException.java
 java/lang/NoClassDefFoundError.java
 java/lang/NoSuchFieldError.java
Index: libraries/javalib/bootstrap.classlist
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/bootstrap.classlist,v
retrieving revision 1.65
diff -u -p -r1.65 bootstrap.classlist
--- libraries/javalib/bootstrap.classlist	8 Dec 2004 08:05:29 -0000	1.65
+++ libraries/javalib/bootstrap.classlist	19 Dec 2004 14:26:32 -0000
@@ -98,7 +98,6 @@ java/lang/InterruptedException.class
 java/lang/LinkageError.class
 java/lang/Long.class
 java/lang/Math.class
-java/lang/NativeLibrary.class
 java/lang/NegativeArraySizeException.class
 java/lang/NoClassDefFoundError.class
 java/lang/NoSuchMethodException.class
Index: kaffe/kaffe/main.c
===================================================================
RCS file: /cvs/kaffe/kaffe/kaffe/kaffe/main.c,v
retrieving revision 1.73
diff -u -p -r1.73 main.c
--- kaffe/kaffe/main.c	18 Dec 2004 07:45:34 -0000	1.73
+++ kaffe/kaffe/main.c	19 Dec 2004 14:26:32 -0000
@@ -198,7 +198,7 @@ main(int argc, char* argv[])
 			libargs += 1;
 		}
 		
-		i = loadNativeLibrary(libpath, errbuf, sizeof(errbuf));
+		i = loadNativeLibrary(libpath, NULL, errbuf, sizeof(errbuf));
 		if( i > 0 )
 		{
 			jint (*onload_func)(JavaVM *jvm, char *, void *);
Index: include/Makefile.am
===================================================================
RCS file: /cvs/kaffe/kaffe/include/Makefile.am,v
retrieving revision 1.79
diff -u -p -r1.79 Makefile.am
--- include/Makefile.am	8 Dec 2004 07:22:07 -0000	1.79
+++ include/Makefile.am	19 Dec 2004 14:26:32 -0000
@@ -67,7 +67,6 @@ NOINSTALL_DERIVED_HDRS = \
 	java_lang_Integer.h \
 	java_lang_Long.h \
 	java_lang_Math.h \
-	java_lang_NativeLibrary.h \
 	java_lang_Runtime.h \
 	java_lang_Short.h \
 	java_lang_System.h \
Index: include/Makefile.in
===================================================================
RCS file: /cvs/kaffe/kaffe/include/Makefile.in,v
retrieving revision 1.197
diff -u -p -r1.197 Makefile.in
--- include/Makefile.in	8 Dec 2004 07:22:07 -0000	1.197
+++ include/Makefile.in	19 Dec 2004 14:26:33 -0000
@@ -416,7 +416,6 @@ NOINSTALL_DERIVED_HDRS = \
 	java_lang_Integer.h \
 	java_lang_Long.h \
 	java_lang_Math.h \
-	java_lang_NativeLibrary.h \
 	java_lang_Runtime.h \
 	java_lang_Short.h \
 	java_lang_System.h \
Index: libraries/javalib/java/lang/Class.java
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/java/lang/Class.java,v
retrieving revision 1.50
diff -u -p -r1.50 Class.java
--- libraries/javalib/java/lang/Class.java	5 Dec 2004 22:43:31 -0000	1.50
+++ libraries/javalib/java/lang/Class.java	19 Dec 2004 14:26:33 -0000
@@ -286,7 +286,7 @@ native public String getName();
 
 public Package getPackage() {
 	ClassLoader loader = getClassLoader();
-	
+
 	if (loader == null) {
 		return PrimordialClassLoader.getSingleton().
 				getPackage(PackageHelper.getPackageName(this));
@@ -295,15 +295,7 @@ public Package getPackage() {
 	return loader.getPackage(PackageHelper.getPackageName(this));
 }
 
-public ProtectionDomain getProtectionDomain() {
-	ClassLoader loader = getClassLoader ();
-
-	if (loader == null) {
-		return PrimordialClassLoader.getSingleton().getProtectionDomain(this);	
-	}
-
-	return loader.getProtectionDomain(this);
-}
+native public ProtectionDomain getProtectionDomain();
 
 native static Class getPrimitiveClass(String name);
 
@@ -492,7 +484,7 @@ public boolean desiredAssertionStatus() 
 	ClassLoader c = getClassLoader();
 	Object status;
 	if (c == null)
-		return ClassLoader.systemDefaultAssertionStatus;
+		return VMClassLoader.defaultAssertionStatus();
 	if (c.classAssertionStatus != null) {
 		synchronized (c) {
 			status = c.classAssertionStatus.get(getName());
@@ -500,7 +492,7 @@ public boolean desiredAssertionStatus() 
 				return status.equals(Boolean.TRUE);
 		}
 	} else {
-		status = ClassLoader.systemClassAssertionStatus.get(getName());
+		status = ClassLoader.StaticData.systemClassAssertionStatus.get(getName());
 		if (status != null)
 			return status.equals(Boolean.TRUE);
 	}
@@ -521,10 +513,10 @@ public boolean desiredAssertionStatus() 
 	} else {
 		String name = getPackagePortion(getName());
 		if ("".equals(name))
-			status = ClassLoader.systemPackageAssertionStatus.get(null);
+			status = ClassLoader.StaticData.systemPackageAssertionStatus.get(null);
 		else {
 			do {
-				status = ClassLoader.systemPackageAssertionStatus.get(name);
+				status = ClassLoader.StaticData.systemPackageAssertionStatus.get(name);
 				name = getPackagePortion(name);
 			} while (! "".equals(name) && status == null);
 		}
Index: libraries/javalib/java/lang/ClassLoader.java
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/java/lang/ClassLoader.java,v
retrieving revision 1.34
diff -u -p -r1.34 ClassLoader.java
--- libraries/javalib/java/lang/ClassLoader.java	5 Dec 2004 22:43:31 -0000	1.34
+++ libraries/javalib/java/lang/ClassLoader.java	19 Dec 2004 14:26:34 -0000
@@ -1,439 +1,1095 @@
-/*
- * Java core library component.
- *
- * Copyright (c) 1997, 1998
- *      Transvirtual Technologies, Inc.  All rights reserved.
- * Portions Copyright (C) 1998, 1999, 2001, 2002, 2003 Free Software Foundation, Inc.
- *
- * See the file "license.terms" for information on usage and redistribution
- * of this file.
- */
+/* ClassLoader.java -- responsible for loading classes into the VM
+   Copyright (C) 1998, 1999, 2001, 2002, 2003, 2004 Free Software Foundation, Inc.
+
+This file is part of GNU Classpath.
+
+GNU Classpath is free software; you can redistribute it and/or modify
+it under the terms of the GNU General Public License as published by
+the Free Software Foundation; either version 2, or (at your option)
+any later version.
+
+GNU Classpath is distributed in the hope that it will be useful, but
+WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with GNU Classpath; see the file COPYING.  If not, write to the
+Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+02111-1307 USA.
+
+Linking this library statically or dynamically with other modules is
+making a combined work based on this library.  Thus, the terms and
+conditions of the GNU General Public License cover the whole
+combination.
+
+As a special exception, the copyright holders of this library give you
+permission to link this library with independent modules to produce an
+executable, regardless of the license terms of these independent
+modules, and to copy and distribute the resulting executable under
+terms of your choice, provided that you also meet, for each linked
+independent module, the terms and conditions of the license of that
+module.  An independent module is a module which is not derived from
+or based on this library.  If you modify this library, you may extend
+this exception to your version of the library, but you are not
+obligated to do so.  If you do not wish to do so, delete this
+exception statement from your version. */
+
 
 package java.lang;
 
+import gnu.classpath.SystemProperties;
+import gnu.java.util.DoubleEnumeration;
+import gnu.java.util.EmptyEnumeration;
+
+import java.io.File;
 import java.io.IOException;
 import java.io.InputStream;
+import java.lang.reflect.Constructor;
 import java.net.URL;
+import java.net.URLClassLoader;
+import java.security.CodeSource;
+import java.security.PermissionCollection;
+import java.security.Policy;
 import java.security.ProtectionDomain;
+import java.util.ArrayList;
 import java.util.Enumeration;
 import java.util.HashMap;
-import java.util.HashSet;
-import java.util.Hashtable;
 import java.util.Map;
-import java.util.NoSuchElementException;
-import java.util.Set;
-import java.util.Vector;
-
-import kaffe.lang.AppClassLoader;
-import kaffe.lang.PrimordialClassLoader;
-
-public abstract class ClassLoader {
-
-/**
- * To prevent any classes from being gc'd before we are gc'd, we keep them
- * in this hashtable.  The contents of this table correspond to entries
- * in the VM-internal class entry pool.  This table only contains classes
- * for which we are the defining loader (ie, not classes for which we are
- * merely the initiating loader).
- */
-private final Hashtable loadedClasses = new Hashtable();
-
-/**
- * Similarly, we keep a reference to all native libraries loaded by this
- * loader.  When this class loader is GC'd, the native libraries become
- * candidates for finalization, which unlinks the shared library.
- */
-private final Set loadedLibraries = new HashSet();
-
-/**
- * We keep a reference to all packages loaded by this loader.
- */
-private final Hashtable loadedPackages = new Hashtable();
-
-/**
- * We keep the references to loaded classes and their protection domain
- * in this Map.
- */
-private final Hashtable protectionDomains = new Hashtable();
-
-/**
- * The command-line state of the default assertion status overrides.
- */
-// Package visible for use by Class.
-// XXX should be set from command line
-static final boolean systemDefaultAssertionStatus = true;
-
-/**
- * The desired assertion status of classes loaded by this loader, if not
- * overridden by package or class instructions.
- */
-// Package visible for use by Class.
-boolean defaultAssertionStatus = systemDefaultAssertionStatus;
-
-/**
- * The command-line state of the package assertion status overrides. This
- * map is never modified, so it does not need to be synchronized.
- */
-// Package visible for use by Class.
-// XXX should be set from command line
-static final Map systemPackageAssertionStatus = new HashMap();
-
-/**
- * The map of package assertion status overrides, or null if no package
- * overrides have been specified yet. The values of the map should be
- * Boolean.TRUE or Boolean.FALSE, and the unnamed package is represented
- * by the null key. This map must be synchronized on this instance.
- */
-// Package visible for use by Class.
-Map packageAssertionStatus;
-
-/**
- * The command-line state of the class assertion status overrides. This
- * map is never modified, so it does not need to be synchronized.
- */
-// Package visible for use by Class.
-// XXX should be set from command line
-static final Map systemClassAssertionStatus = new HashMap();
-
-/**
- * The map of class assertion status overrides, or null if no class
- * overrides have been specified yet. The values of the map should be
- * Boolean.TRUE or Boolean.FALSE. This map must be synchronized on this
- * instance.
- */
-// Package visible for use by Class.
-Map classAssertionStatus;
-
-private ProtectionDomain defaultProtectionDomain;
-private final ClassLoader parent;
-
-protected ClassLoader() {
-	this(getSystemClassLoader());
-}
-
-protected ClassLoader(ClassLoader parent) {
-	SecurityManager sm = System.getSecurityManager();
-	if (sm != null)
-		sm.checkCreateClassLoader();
-	this.parent = parent;
-}
-
-public Class loadClass(String name) throws ClassNotFoundException {
-	return (loadClass(name, false));
-}
-
-protected Class loadClass(String name, boolean resolve)
-		throws ClassNotFoundException {
-	Class c = findLoadedClass (name);
-
-	// Second, try the parent class loader
-	if (c == null) {
-		try {
-			if (parent != null) {
-				c = parent.loadClass(name, resolve);
-			} else {
-				c = PrimordialClassLoader.getSingleton().
-					loadClass(name, resolve);
-			}
-		} catch (ClassNotFoundException _) {
-		}
-	}
-
-	// Third, try findClass()
-	if (c==null && (c = findClass(name)) == null) {
-		throw new ClassNotFoundException(name);
-	}
-
-	// Now, optionally resolve the class
-	if (resolve) {
-		resolveClass(c);
-	}
-	return (c);
-}
-
-protected Class findClass(String name) throws ClassNotFoundException {
-	throw new ClassNotFoundException(name);
-}
+import java.util.StringTokenizer;
 
 /**
- * @deprecated
- */
-protected final Class defineClass(byte data[], int off, int len)
-		throws ClassFormatError {
-	return (defineClass(null, data, off, len));
-}
-
-protected final Class defineClass(String name, byte data[], int off, int len)
-		throws ClassFormatError {
-	return defineClass(name, data, off, len, null);
-}
-
-protected final Class defineClass(String name, byte data[], int off,
-		int len, ProtectionDomain pd) throws ClassFormatError {
-	if (off < 0 || len < 0 || off + len > data.length) {
-		throw new IndexOutOfBoundsException();
-	}
-	Class clazz = null;
-	clazz = defineClass0(name, data, off, len);
-	if (name != null) {
-		loadedClasses.put(name, clazz);
-	}
-	else {
-		loadedClasses.put(clazz.getName(), clazz);
-	}
-	if (pd == null) {
-		if (defaultProtectionDomain == null) {
-			// XXX FIXME..
-			defaultProtectionDomain = new ProtectionDomain(null, null);
-		}
-		pd = defaultProtectionDomain;
-	}
-	protectionDomains.put(clazz, pd);
-	return (clazz);
-}
-
-protected final void resolveClass(Class c) {
-	resolveClass0(c);
-}
-
-protected final Class findSystemClass(String name)
-		throws ClassNotFoundException {
-	return getSystemClassLoader().loadClass(name);
-}
-
-public final ClassLoader getParent() {
-	return parent;
-}
-
-protected final void setSigners(Class cl, Object[] signers) {
-        cl.setSigners (signers);
-}
-
-protected final Class findLoadedClass(String name) {
-	return (Class) loadedClasses.get(name); 
-}
-
-public URL getResource(String name) {
-	URL ret = null;
-
-	if (parent != null) {
-		ret = parent.getResource (name);
-	} else {
-		// we have to call findResource directly, since getResource()
-		// would cause an endless loop 
-		ret = PrimordialClassLoader.getSingleton().findResource (name);
-	}
-
-	if (ret == null) {
-		ret = findResource (name);
-	}
-	
-	return ret;
-}
-
-public final Enumeration getResources(String name) throws IOException {
-	Vector v = new Vector();
-	
-	if (parent != null) {
-		for (Enumeration e = parent.getResources(name); e.hasMoreElements();) {
-			v.addElement(e.nextElement());
-		}
-	} else {
-		// we have to call findResources directly, since getResources() would cause an endless loop
-		for (Enumeration e=PrimordialClassLoader.getSingleton().findResources(name); e.hasMoreElements();) {
-			v.addElement (e.nextElement ());
-		}
-	}
-
-	for (Enumeration e = findResources(name); e.hasMoreElements(); )
-		v.addElement(e.nextElement());
-
-	return v.elements();
-}
-
-protected Enumeration findResources(String name) throws IOException {
-	return new Vector().elements();		// ie, an empty Enumeration
-}
-
-protected URL findResource(String name) {
-	try {
-		return (URL)findResources(name).nextElement();
-	} catch (IOException e) {
-	} catch (NoSuchElementException e) {
-	}
-	return null;
-}
-
-public static URL getSystemResource(String name) {
-	return getSystemClassLoader().getResource(name);
-}
-
-public static Enumeration getSystemResources(String name) throws IOException {
-	return getSystemClassLoader().getResources(name);
-}
-
-public InputStream getResourceAsStream(String name) {
-	URL url = getResource(name);
-
-	if (url != null) {
-		try {
-			return url.openStream();
-		} catch (IOException e) {
-		}
-	}
-	return null;
-}
-
-public static InputStream getSystemResourceAsStream(String name) {
-	return getSystemClassLoader().getResourceAsStream(name);
-}
-
-public static ClassLoader getSystemClassLoader() {
-	return AppClassLoader.getSingleton();
-}
-
-protected Package definePackage(String name, String specTitle,
-				String specVersion, String specVendor,
-				String implTitle, String implVersion,
-				String implVendor, URL sealBase)
-	throws IllegalArgumentException {
-
-	Package pack = getPackage(name);
-
-	if (pack != null) {
-		throw new IllegalArgumentException("Package " + name + " already defined");
-	}
-
-	pack = new Package(name, specTitle, specVersion, specVendor,
-			   implTitle, implVersion, implVendor, sealBase);
-	loadedPackages.put(name, pack);
-	return pack;
-}
-
-protected Package getPackage(String name) {
-	Package ret = null;
-
-	if (parent!=null) {
-		ret = parent.getPackage (name);
-	} else {
-		ret = PrimordialClassLoader.getSingleton().getPackage(name);
-	}	
-
-	if (ret == null) {
-		ret = (Package) loadedPackages.get (name);
-	}
-
-	return ret;
-}
-
-protected Package[] getPackages() {
-	Package[] myPackages;
-
-	myPackages = new Package[loadedPackages.size()];
-	loadedPackages.values().toArray (myPackages);
-
-	Package[] ancestorPackages = null;
-
-	if (parent != null) {
-		ancestorPackages = parent.getPackages();
-	} else {
-		ancestorPackages = PrimordialClassLoader.getSingleton().getPackages();
-	} 
-	
-	Package[] ret = new Package[myPackages.length + ancestorPackages.length];
-	
-	System.arraycopy (ret, 0, ancestorPackages, 0, ancestorPackages.length);
-	System.arraycopy (ret, ancestorPackages.length, myPackages, 0, myPackages.length);
-		
-	return ret;
-}
-
-protected String findLibrary(String libname) {
-	return null;
-}
-
-synchronized void addNativeLibrary(NativeLibrary lib) {
-	loadedLibraries.add(lib);
-}
-
-ProtectionDomain getProtectionDomain(Class clazz) {
-	return (ProtectionDomain) protectionDomains.get(clazz);
-}
-
-private native Class defineClass0(String name, byte data[], int off, int len);
-private native void resolveClass0(Class cls);
-
-/**
- * Set the default assertion status for classes loaded by this classloader,
- * used unless overridden by a package or class request.
+ * The ClassLoader is a way of customizing the way Java gets its classes
+ * and loads them into memory.  The verifier and other standard Java things
+ * still run, but the ClassLoader is allowed great flexibility in determining
+ * where to get the classfiles and when to load and resolve them. For that
+ * matter, a custom ClassLoader can perform on-the-fly code generation or
+ * modification!
  *
- * @param enabled true to set the default to enabled
- * @see #setClassAssertionStatus(String, boolean)
- * @see #setPackageAssertionStatus(String, boolean)
- * @see #clearAssertionStatus()
- * @since 1.4
- */
-public void setDefaultAssertionStatus(boolean enabled)
-{
-	defaultAssertionStatus = enabled;
-}
-  
-/**
- * Set the default assertion status for packages, used unless overridden
- * by a class request. This default also covers subpackages, unless they
- * are also specified. The unnamed package should use null for the name.
+ * <p>Every classloader has a parent classloader that is consulted before
+ * the 'child' classloader when classes or resources should be loaded.
+ * This is done to make sure that classes can be loaded from an hierarchy of
+ * multiple classloaders and classloaders do not accidentially redefine
+ * already loaded classes by classloaders higher in the hierarchy.
  *
- * @param name the package (and subpackages) to affect
- * @param enabled true to set the default to enabled
- * @see #setDefaultAssertionStatus(String, boolean)
- * @see #setClassAssertionStatus(String, boolean)
- * @see #clearAssertionStatus()
- * @since 1.4
- */
-public synchronized void setPackageAssertionStatus(String name,
-                                                   boolean enabled)
-{
-	if (packageAssertionStatus == null)
-		packageAssertionStatus
-		  = new HashMap(systemPackageAssertionStatus);
-	packageAssertionStatus.put(name, enabled ? Boolean.TRUE : Boolean.FALSE);
-}
-  
-/**
- * Set the default assertion status for a class. This only affects the
- * status of top-level classes, any other string is harmless.
+ * <p>The grandparent of all classloaders is the bootstrap classloader, which
+ * loads all the standard system classes as implemented by GNU Classpath. The
+ * other special classloader is the system classloader (also called
+ * application classloader) that loads all classes from the CLASSPATH
+ * (<code>java.class.path</code> system property). The system classloader
+ * is responsible for finding the application classes from the classpath,
+ * and delegates all requests for the standard library classes to its parent
+ * the bootstrap classloader. Most programs will load all their classes
+ * through the system classloaders.
  *
- * @param name the class to affect
- * @param enabled true to set the default to enabled
- * @throws NullPointerException if name is null
- * @see #setDefaultAssertionStatus(String, boolean)
- * @see #setPackageAssertionStatus(String, boolean)
- * @see #clearAssertionStatus()
- * @since 1.4
- */
-public synchronized void setClassAssertionStatus(String name,
-                                                 boolean enabled)
-{
-	if (classAssertionStatus == null)
-		classAssertionStatus = new HashMap(systemClassAssertionStatus);
-	// The toString() hack catches null, as required.
-	classAssertionStatus.put(name.toString(), enabled ? Boolean.TRUE : Boolean.FALSE);
-}
-  
-/**
- * Resets the default assertion status of this classloader, its packages
- * and classes, all to false. This allows overriding defaults inherited
- * from the command line.
+ * <p>The bootstrap classloader in GNU Classpath is implemented as a couple of
+ * static (native) methods on the package private class
+ * <code>java.lang.VMClassLoader</code>, the system classloader is an
+ * anonymous inner class of ClassLoader and a subclass of
+ * <code>java.net.URLClassLoader</code>.
+ *
+ * <p>Users of a <code>ClassLoader</code> will normally just use the methods
+ * <ul>
+ *  <li> <code>loadClass()</code> to load a class.</li>
+ *  <li> <code>getResource()</code> or <code>getResourceAsStream()</code>
+ *       to access a resource.</li>
+ *  <li> <code>getResources()</code> to get an Enumeration of URLs to all
+ *       the resources provided by the classloader and its parents with the
+ *       same name.</li>
+ * </ul>
  *
- * @see #setDefaultAssertionStatus(boolean)
- * @see #setClassAssertionStatus(String, boolean)
- * @see #setPackageAssertionStatus(String, boolean)
- * @since 1.4
+ * <p>Subclasses should implement the methods
+ * <ul>
+ *  <li> <code>findClass()</code> which is called by <code>loadClass()</code>
+ *       when the parent classloader cannot provide a named class.</li>
+ *  <li> <code>findResource()</code> which is called by
+ *       <code>getResource()</code> when the parent classloader cannot provide
+ *       a named resource.</li>
+ *  <li> <code>findResources()</code> which is called by
+ *       <code>getResource()</code> to combine all the resources with the
+ *       same name from the classloader and its parents.</li>
+ *  <li> <code>findLibrary()</code> which is called by
+ *       <code>Runtime.loadLibrary()</code> when a class defined by the
+ *       classloader wants to load a native library.</li>
+ * </ul>
+ *
+ * @author John Keiser
+ * @author Mark Wielaard
+ * @author Eric Blake <ebb9@email.byu.edu>
+ * @see Class
+ * @since 1.0
+ * @status still missing 1.4 functionality
  */
-public synchronized void clearAssertionStatus()
+public abstract class ClassLoader
 {
-	defaultAssertionStatus = false;
-	packageAssertionStatus = new HashMap();
-	classAssertionStatus = new HashMap();
-}
+  /**
+   * All classes loaded by this classloader. VM's may choose to implement
+   * this cache natively; but it is here available for use if necessary. It
+   * is not private in order to allow native code (and trusted subclasses)
+   * access to this field.
+   */
+  final HashMap loadedClasses = new HashMap();
+
+  /**
+   * All packages defined by this classloader. It is not private in order to
+   * allow native code (and trusted subclasses) access to this field.
+   */
+  final HashMap definedPackages = new HashMap();
+
+  /**
+   * The classloader that is consulted before this classloader.
+   * If null then the parent is the bootstrap classloader.
+   */
+  private final ClassLoader parent;
+
+  /**
+   * This is true if this classloader was successfully initialized.
+   * This flag is needed to avoid a class loader attack: even if the
+   * security manager rejects an attempt to create a class loader, the
+   * malicious class could have a finalize method which proceeds to
+   * define classes.
+   */
+  private final boolean initialized;
+
+  static class StaticData
+  {
+    /**
+     * The System Class Loader (a.k.a. Application Class Loader). The one
+     * returned by ClassLoader.getSystemClassLoader.
+     */
+    static final ClassLoader systemClassLoader =
+                              VMClassLoader.getSystemClassLoader();
+    static
+    {
+      // Find out if we have to install a default security manager. Note that
+      // this is done here because we potentially need the system class loader
+      // to load the security manager and note also that we don't need the
+      // security manager until the system class loader is created.
+      // If the runtime chooses to use a class loader that doesn't have the
+      // system class loader as its parent, it is responsible for setting
+      // up a security manager before doing so.
+      String secman = SystemProperties.getProperty("java.security.manager");
+      if (secman != null && SecurityManager.current == null)
+        {
+          if (secman.equals("") || secman.equals("default"))
+	    {
+	      SecurityManager.current = new SecurityManager();
+	    }
+	  else
+	    {
+	      try
+	        {
+	  	  Class cl = Class.forName(secman, false, StaticData.systemClassLoader);
+		  SecurityManager.current = (SecurityManager)cl.newInstance();
+	        }
+	      catch (Exception x)
+	        {
+		  throw (InternalError)
+		      new InternalError("Unable to create SecurityManager")
+		  	  .initCause(x);
+	        }
+	    }
+        }
+    }
+
+    /**
+     * The default protection domain, used when defining a class with a null
+     * parameter for the domain.
+     */
+    static final ProtectionDomain defaultProtectionDomain;
+    static
+    {
+        CodeSource cs = new CodeSource(null, null);
+        PermissionCollection perm = Policy.getPolicy().getPermissions(cs);
+        defaultProtectionDomain = new ProtectionDomain(cs, perm);
+    }
+    /**
+     * The command-line state of the package assertion status overrides. This
+     * map is never modified, so it does not need to be synchronized.
+     */
+    // Package visible for use by Class.
+    static final Map systemPackageAssertionStatus
+      = VMClassLoader.packageAssertionStatus();
+    /**
+     * The command-line state of the class assertion status overrides. This
+     * map is never modified, so it does not need to be synchronized.
+     */
+    // Package visible for use by Class.
+    static final Map systemClassAssertionStatus
+      = VMClassLoader.classAssertionStatus();
+  }
+
+  /**
+   * The desired assertion status of classes loaded by this loader, if not
+   * overridden by package or class instructions.
+   */
+  // Package visible for use by Class.
+  boolean defaultAssertionStatus = VMClassLoader.defaultAssertionStatus();
+
+  /**
+   * The map of package assertion status overrides, or null if no package
+   * overrides have been specified yet. The values of the map should be
+   * Boolean.TRUE or Boolean.FALSE, and the unnamed package is represented
+   * by the null key. This map must be synchronized on this instance.
+   */
+  // Package visible for use by Class.
+  Map packageAssertionStatus;
+
+  /**
+   * The map of class assertion status overrides, or null if no class
+   * overrides have been specified yet. The values of the map should be
+   * Boolean.TRUE or Boolean.FALSE. This map must be synchronized on this
+   * instance.
+   */
+  // Package visible for use by Class.
+  Map classAssertionStatus;
+
+  /**
+   * Create a new ClassLoader with as parent the system classloader. There
+   * may be a security check for <code>checkCreateClassLoader</code>.
+   *
+   * @throws SecurityException if the security check fails
+   */
+  protected ClassLoader() throws SecurityException
+  {
+    this(StaticData.systemClassLoader);
+  }
+
+  /**
+   * Create a new ClassLoader with the specified parent. The parent will
+   * be consulted when a class or resource is requested through
+   * <code>loadClass()</code> or <code>getResource()</code>. Only when the
+   * parent classloader cannot provide the requested class or resource the
+   * <code>findClass()</code> or <code>findResource()</code> method
+   * of this classloader will be called. There may be a security check for
+   * <code>checkCreateClassLoader</code>.
+   *
+   * @param parent the classloader's parent, or null for the bootstrap
+   *        classloader
+   * @throws SecurityException if the security check fails
+   * @since 1.2
+   */
+  protected ClassLoader(ClassLoader parent)
+  {
+    // May we create a new classloader?
+    SecurityManager sm = SecurityManager.current;
+    if (sm != null)
+      sm.checkCreateClassLoader();
+    this.parent = parent;
+    this.initialized = true;
+  }
+
+  /**
+   * Load a class using this ClassLoader or its parent, without resolving
+   * it. Calls <code>loadClass(name, false)</code>.
+   *
+   * <p>Subclasses should not override this method but should override
+   * <code>findClass()</code> which is called by this method.</p>
+   *
+   * @param name the name of the class relative to this ClassLoader
+   * @return the loaded class
+   * @throws ClassNotFoundException if the class cannot be found
+   */
+  public Class loadClass(String name) throws ClassNotFoundException
+  {
+    return loadClass(name, false);
+  }
+
+  /**
+   * Load a class using this ClassLoader or its parent, possibly resolving
+   * it as well using <code>resolveClass()</code>. It first tries to find
+   * out if the class has already been loaded through this classloader by
+   * calling <code>findLoadedClass()</code>. Then it calls
+   * <code>loadClass()</code> on the parent classloader (or when there is
+   * no parent it uses the VM bootclassloader). If the class is still
+   * not loaded it tries to create a new class by calling
+   * <code>findClass()</code>. Finally when <code>resolve</code> is
+   * <code>true</code> it also calls <code>resolveClass()</code> on the
+   * newly loaded class.
+   *
+   * <p>Subclasses should not override this method but should override
+   * <code>findClass()</code> which is called by this method.</p>
+   *
+   * @param name the fully qualified name of the class to load
+   * @param resolve whether or not to resolve the class
+   * @return the loaded class
+   * @throws ClassNotFoundException if the class cannot be found
+   */
+  protected synchronized Class loadClass(String name, boolean resolve)
+    throws ClassNotFoundException
+  {
+    // Have we already loaded this class?
+    Class c = findLoadedClass(name);
+    if (c != null)
+      return c;
+
+    // Can the class be loaded by a parent?
+    try
+      {
+	if (parent == null)
+	  {
+	    c = VMClassLoader.loadClass(name, resolve);
+	    if (c != null)
+	      return c;
+	  }
+	else
+	  {
+	    return parent.loadClass(name, resolve);
+	  }
+      }
+    catch (ClassNotFoundException e)
+      {
+      }
+    // Still not found, we have to do it ourself.
+    c = findClass(name);
+    if (resolve)
+      resolveClass(c);
+    return c;
+  }
+
+  /**
+   * Called for every class name that is needed but has not yet been
+   * defined by this classloader or one of its parents. It is called by
+   * <code>loadClass()</code> after both <code>findLoadedClass()</code> and
+   * <code>parent.loadClass()</code> couldn't provide the requested class.
+   *
+   * <p>The default implementation throws a
+   * <code>ClassNotFoundException</code>. Subclasses should override this
+   * method. An implementation of this method in a subclass should get the
+   * class bytes of the class (if it can find them), if the package of the
+   * requested class doesn't exist it should define the package and finally
+   * it should call define the actual class. It does not have to resolve the
+   * class. It should look something like the following:<br>
+   *
+   * <pre>
+   * // Get the bytes that describe the requested class
+   * byte[] classBytes = classLoaderSpecificWayToFindClassBytes(name);
+   * // Get the package name
+   * int lastDot = name.lastIndexOf('.');
+   * if (lastDot != -1)
+   *   {
+   *     String packageName = name.substring(0, lastDot);
+   *     // Look if the package already exists
+   *     if (getPackage(pkg) == null)
+   *       {
+   *         // define the package
+   *         definePackage(packageName, ...);
+   *       }
+   *   }
+   * // Define and return the class
+   *  return defineClass(name, classBytes, 0, classBytes.length);
+   * </pre>
+   *
+   * <p><code>loadClass()</code> makes sure that the <code>Class</code>
+   * returned by <code>findClass()</code> will later be returned by
+   * <code>findLoadedClass()</code> when the same class name is requested.
+   *
+   * @param name class name to find (including the package name)
+   * @return the requested Class
+   * @throws ClassNotFoundException when the class can not be found
+   * @since 1.2
+   */
+  protected Class findClass(String name) throws ClassNotFoundException
+  {
+    throw new ClassNotFoundException(name);
+  }
+
+  /**
+   * Helper to define a class using a string of bytes. This version is not
+   * secure.
+   *
+   * @param data the data representing the classfile, in classfile format
+   * @param offset the offset into the data where the classfile starts
+   * @param len the length of the classfile data in the array
+   * @return the class that was defined
+   * @throws ClassFormatError if data is not in proper classfile format
+   * @throws IndexOutOfBoundsException if offset or len is negative, or
+   *         offset + len exceeds data
+   * @deprecated use {@link #defineClass(String, byte[], int, int)} instead
+   */
+  protected final Class defineClass(byte[] data, int offset, int len)
+    throws ClassFormatError
+  {
+    return defineClass(null, data, offset, len);
+  }
+
+  /**
+   * Helper to define a class using a string of bytes without a
+   * ProtectionDomain. Subclasses should call this method from their
+   * <code>findClass()</code> implementation. The name should use '.'
+   * separators, and discard the trailing ".class".  The default protection
+   * domain has the permissions of
+   * <code>Policy.getPolicy().getPermissions(new CodeSource(null, null))</code>.
+   *
+   * @param name the name to give the class, or null if unknown
+   * @param data the data representing the classfile, in classfile format
+   * @param offset the offset into the data where the classfile starts
+   * @param len the length of the classfile data in the array
+   * @return the class that was defined
+   * @throws ClassFormatError if data is not in proper classfile format
+   * @throws IndexOutOfBoundsException if offset or len is negative, or
+   *         offset + len exceeds data
+   * @throws SecurityException if name starts with "java."
+   * @since 1.1
+   */
+  protected final Class defineClass(String name, byte[] data, int offset,
+                                    int len) throws ClassFormatError
+  {
+    return defineClass(name, data, offset, len, null);
+  }
+
+  /**
+   * Helper to define a class using a string of bytes. Subclasses should call
+   * this method from their <code>findClass()</code> implementation. If the
+   * domain is null, the default of
+   * <code>Policy.getPolicy().getPermissions(new CodeSource(null, null))</code>
+   * is used. Once a class has been defined in a package, all further classes
+   * in that package must have the same set of certificates or a
+   * SecurityException is thrown.
+   *
+   * @param name the name to give the class.  null if unknown
+   * @param data the data representing the classfile, in classfile format
+   * @param offset the offset into the data where the classfile starts
+   * @param len the length of the classfile data in the array
+   * @param domain the ProtectionDomain to give to the class, null for the
+   *        default protection domain
+   * @return the class that was defined
+   * @throws ClassFormatError if data is not in proper classfile format
+   * @throws IndexOutOfBoundsException if offset or len is negative, or
+   *         offset + len exceeds data
+   * @throws SecurityException if name starts with "java.", or if certificates
+   *         do not match up
+   * @since 1.2
+   */
+  protected final synchronized Class defineClass(String name, byte[] data,
+						 int offset, int len,
+						 ProtectionDomain domain)
+    throws ClassFormatError
+  {
+    if (domain == null)
+      domain = StaticData.defaultProtectionDomain;
+    if (! initialized)
+      throw new SecurityException("attempt to define class from uninitialized class loader");
+    Class retval = VMClassLoader.defineClass(this, name, data,
+                                             offset, len, domain);
+    loadedClasses.put(retval.getName(), retval);
+    return retval;
+  }
+
+  /**
+   * Links the class, if that has not already been done. Linking basically
+   * resolves all references to other classes made by this class.
+   *
+   * @param c the class to resolve
+   * @throws NullPointerException if c is null
+   * @throws LinkageError if linking fails
+   */
+  protected final void resolveClass(Class c)
+  {
+    VMClassLoader.resolveClass(c);
+  }
+
+  /**
+   * Helper to find a Class using the system classloader, possibly loading it.
+   * A subclass usually does not need to call this, if it correctly
+   * overrides <code>findClass(String)</code>.
+   *
+   * @param name the name of the class to find
+   * @return the found class
+   * @throws ClassNotFoundException if the class cannot be found
+   */
+  protected final Class findSystemClass(String name)
+    throws ClassNotFoundException
+  {
+    return Class.forName(name, false, StaticData.systemClassLoader);
+  }
+
+  /**
+   * Returns the parent of this classloader. If the parent of this
+   * classloader is the bootstrap classloader then this method returns
+   * <code>null</code>. A security check may be performed on
+   * <code>RuntimePermission("getClassLoader")</code>.
+   *
+   * @return the parent <code>ClassLoader</code>
+   * @throws SecurityException if the security check fails
+   * @since 1.2
+   */
+  public final ClassLoader getParent()
+  {
+    // Check if we may return the parent classloader.
+    SecurityManager sm = SecurityManager.current;
+    if (sm != null)
+      {
+        Class c = VMSecurityManager.getClassContext()[1];
+        ClassLoader cl = c.getClassLoader();
+	if (cl != null && ! cl.isAncestorOf(this))
+          sm.checkPermission(new RuntimePermission("getClassLoader"));
+      }
+    return parent;
+  }
+
+  /**
+   * Helper to set the signers of a class. This should be called after
+   * defining the class.
+   *
+   * @param c the Class to set signers of
+   * @param signers the signers to set
+   * @since 1.1
+   */
+  protected final void setSigners(Class c, Object[] signers)
+  {
+    c.setSigners(signers);
+  }
+
+  /**
+   * Helper to find an already-loaded class in this ClassLoader.
+   *
+   * @param name the name of the class to find
+   * @return the found Class, or null if it is not found
+   * @since 1.1
+   */
+  protected final synchronized Class findLoadedClass(String name)
+  {
+    // NOTE: If the VM is keeping its own cache, it may make sense to have
+    // this method be native.
+    return (Class) loadedClasses.get(name);
+  }
+
+  /**
+   * Get the URL to a resource using this classloader or one of its parents.
+   * First tries to get the resource by calling <code>getResource()</code>
+   * on the parent classloader. If the parent classloader returns null then
+   * it tries finding the resource by calling <code>findResource()</code> on
+   * this classloader. The resource name should be separated by '/' for path
+   * elements.
+   *
+   * <p>Subclasses should not override this method but should override
+   * <code>findResource()</code> which is called by this method.
+   *
+   * @param name the name of the resource relative to this classloader
+   * @return the URL to the resource or null when not found
+   */
+  public URL getResource(String name)
+  {
+    URL result;
+
+    if (parent == null)
+      result = VMClassLoader.getResource(name);
+    else
+      result = parent.getResource(name);
+
+    if (result == null)
+      result = findResource(name);
+    return result;
+  }
+
+  /**
+   * Returns an Enumeration of all resources with a given name that can
+   * be found by this classloader and its parents. Certain classloaders
+   * (such as the URLClassLoader when given multiple jar files) can have
+   * multiple resources with the same name that come from multiple locations.
+   * It can also occur that a parent classloader offers a resource with a
+   * certain name and the child classloader also offers a resource with that
+   * same name. <code>getResource()</code> only offers the first resource (of the
+   * parent) with a given name. This method lists all resources with the
+   * same name. The name should use '/' as path separators.
+   *
+   * <p>The Enumeration is created by first calling <code>getResources()</code>
+   * on the parent classloader and then calling <code>findResources()</code>
+   * on this classloader.</p>
+   *
+   * @param name the resource name
+   * @return an enumaration of all resources found
+   * @throws IOException if I/O errors occur in the process
+   * @since 1.2
+   */
+  public final Enumeration getResources(String name) throws IOException
+  {
+    Enumeration parentResources;
+    if (parent == null)
+      parentResources = VMClassLoader.getResources(name);
+    else
+      parentResources = parent.getResources(name);
+    return new DoubleEnumeration(parentResources, findResources(name));
+  }
+
+  /**
+   * Called whenever all locations of a named resource are needed.
+   * It is called by <code>getResources()</code> after it has called
+   * <code>parent.getResources()</code>. The results are combined by
+   * the <code>getResources()</code> method.
+   *
+   * <p>The default implementation always returns an empty Enumeration.
+   * Subclasses should override it when they can provide an Enumeration of
+   * URLs (possibly just one element) to the named resource.
+   * The first URL of the Enumeration should be the same as the one
+   * returned by <code>findResource</code>.
+   *
+   * @param name the name of the resource to be found
+   * @return a possibly empty Enumeration of URLs to the named resource
+   * @throws IOException if I/O errors occur in the process
+   * @since 1.2
+   */
+  protected Enumeration findResources(String name) throws IOException
+  {
+    return EmptyEnumeration.getInstance();
+  }
+
+  /**
+   * Called whenever a resource is needed that could not be provided by
+   * one of the parents of this classloader. It is called by
+   * <code>getResource()</code> after <code>parent.getResource()</code>
+   * couldn't provide the requested resource.
+   *
+   * <p>The default implementation always returns null. Subclasses should
+   * override this method when they can provide a way to return a URL
+   * to a named resource.
+   *
+   * @param name the name of the resource to be found
+   * @return a URL to the named resource or null when not found
+   * @since 1.2
+   */
+  protected URL findResource(String name)
+  {
+    return null;
+  }
+
+  /**
+   * Get the URL to a resource using the system classloader.
+   *
+   * @param name the name of the resource relative to the system classloader
+   * @return the URL to the resource
+   * @since 1.1
+   */
+  public static final URL getSystemResource(String name)
+  {
+    return StaticData.systemClassLoader.getResource(name);
+  }
+
+  /**
+   * Get an Enumeration of URLs to resources with a given name using the
+   * the system classloader. The enumeration firsts lists the resources with
+   * the given name that can be found by the bootstrap classloader followed
+   * by the resources with the given name that can be found on the classpath.
+   *
+   * @param name the name of the resource relative to the system classloader
+   * @return an Enumeration of URLs to the resources
+   * @throws IOException if I/O errors occur in the process
+   * @since 1.2
+   */
+  public static Enumeration getSystemResources(String name) throws IOException
+  {
+    return StaticData.systemClassLoader.getResources(name);
+  }
+
+  /**
+   * Get a resource as stream using this classloader or one of its parents.
+   * First calls <code>getResource()</code> and if that returns a URL to
+   * the resource then it calls and returns the InputStream given by
+   * <code>URL.openStream()</code>.
+   *
+   * <p>Subclasses should not override this method but should override
+   * <code>findResource()</code> which is called by this method.
+   *
+   * @param name the name of the resource relative to this classloader
+   * @return an InputStream to the resource, or null
+   * @since 1.1
+   */
+  public InputStream getResourceAsStream(String name)
+  {
+    try
+      {
+        URL url = getResource(name);
+        if (url == null)
+          return null;
+        return url.openStream();
+      }
+    catch (IOException e)
+      {
+        return null;
+      }
+  }
+
+  /**
+   * Get a resource using the system classloader.
+   *
+   * @param name the name of the resource relative to the system classloader
+   * @return an input stream for the resource, or null
+   * @since 1.1
+   */
+  public static final InputStream getSystemResourceAsStream(String name)
+  {
+    try
+      {
+        URL url = getSystemResource(name);
+        if (url == null)
+          return null;
+        return url.openStream();
+      }
+    catch (IOException e)
+      {
+        return null;
+      }
+  }
+
+  /**
+   * Returns the system classloader. The system classloader (also called
+   * the application classloader) is the classloader that is used to
+   * load the application classes on the classpath (given by the system
+   * property <code>java.class.path</code>. This is set as the context
+   * class loader for a thread. The system property
+   * <code>java.system.class.loader</code>, if defined, is taken to be the
+   * name of the class to use as the system class loader, which must have
+   * a public constructor which takes a ClassLoader as a parent. The parent
+   * class loader passed in the constructor is the default system class
+   * loader.
+   *
+   * <p>Note that this is different from the bootstrap classloader that
+   * actually loads all the real "system" classes.
+   *
+   * <p>A security check will be performed for
+   * <code>RuntimePermission("getClassLoader")</code> if the calling class
+   * is not a parent of the system class loader.
+   *
+   * @return the system class loader
+   * @throws SecurityException if the security check fails
+   * @throws IllegalStateException if this is called recursively
+   * @throws Error if <code>java.system.class.loader</code> fails to load
+   * @since 1.2
+   */
+  public static ClassLoader getSystemClassLoader()
+  {
+    // Check if we may return the system classloader
+    SecurityManager sm = SecurityManager.current;
+    if (sm != null)
+      {
+	Class c = VMSecurityManager.getClassContext()[1];
+	ClassLoader cl = c.getClassLoader();
+	if (cl != null && cl != StaticData.systemClassLoader)
+	  sm.checkPermission(new RuntimePermission("getClassLoader"));
+      }
+
+    return StaticData.systemClassLoader;
+  }
+
+  /**
+   * Defines a new package and creates a Package object. The package should
+   * be defined before any class in the package is defined with
+   * <code>defineClass()</code>. The package should not yet be defined
+   * before in this classloader or in one of its parents (which means that
+   * <code>getPackage()</code> should return <code>null</code>). All
+   * parameters except the <code>name</code> of the package may be
+   * <code>null</code>.
+   *
+   * <p>Subclasses should call this method from their <code>findClass()</code>
+   * implementation before calling <code>defineClass()</code> on a Class
+   * in a not yet defined Package (which can be checked by calling
+   * <code>getPackage()</code>).
+   *
+   * @param name the name of the Package
+   * @param specTitle the name of the specification
+   * @param specVendor the name of the specification designer
+   * @param specVersion the version of this specification
+   * @param implTitle the name of the implementation
+   * @param implVendor the vendor that wrote this implementation
+   * @param implVersion the version of this implementation
+   * @param sealed if sealed the origin of the package classes
+   * @return the Package object for the specified package
+   * @throws IllegalArgumentException if the package name is null or it
+   *         was already defined by this classloader or one of its parents
+   * @see Package
+   * @since 1.2
+   */
+  protected Package definePackage(String name, String specTitle,
+                                  String specVendor, String specVersion,
+                                  String implTitle, String implVendor,
+                                  String implVersion, URL sealed)
+  {
+    if (getPackage(name) != null)
+      throw new IllegalArgumentException("Package " + name
+                                         + " already defined");
+    Package p = new Package(name, specTitle, specVendor, specVersion,
+                            implTitle, implVendor, implVersion, sealed);
+    synchronized (definedPackages)
+      {
+        definedPackages.put(name, p);
+      }
+    return p;
+  }
+
+  /**
+   * Returns the Package object for the requested package name. It returns
+   * null when the package is not defined by this classloader or one of its
+   * parents.
+   *
+   * @param name the package name to find
+   * @return the package, if defined
+   * @since 1.2
+   */
+  protected Package getPackage(String name)
+  {
+    Package p;
+    if (parent == null)
+      p = VMClassLoader.getPackage(name);
+    else
+      p = parent.getPackage(name);
+
+    if (p == null)
+      {
+	synchronized (definedPackages)
+	  {
+	    p = (Package) definedPackages.get(name);
+	  }
+      }
+    return p;
+  }
+
+  /**
+   * Returns all Package objects defined by this classloader and its parents.
+   *
+   * @return an array of all defined packages
+   * @since 1.2
+   */
+  protected Package[] getPackages()
+  {
+    // Get all our packages.
+    Package[] packages;
+    synchronized(definedPackages)
+      {
+        packages = new Package[definedPackages.size()];
+        definedPackages.values().toArray(packages);
+      }
+
+    // If we have a parent get all packages defined by our parents.
+    Package[] parentPackages;
+    if (parent == null)
+      parentPackages = VMClassLoader.getPackages();
+    else
+      parentPackages = parent.getPackages();
+
+    Package[] allPackages = new Package[parentPackages.length
+					+ packages.length];
+    System.arraycopy(parentPackages, 0, allPackages, 0,
+                     parentPackages.length);
+    System.arraycopy(packages, 0, allPackages, parentPackages.length,
+                     packages.length);
+    return allPackages;
+  }
+
+  /**
+   * Called by <code>Runtime.loadLibrary()</code> to get an absolute path
+   * to a (system specific) library that was requested by a class loaded
+   * by this classloader. The default implementation returns
+   * <code>null</code>. It should be implemented by subclasses when they
+   * have a way to find the absolute path to a library. If this method
+   * returns null the library is searched for in the default locations
+   * (the directories listed in the <code>java.library.path</code> system
+   * property).
+   *
+   * @param name the (system specific) name of the requested library
+   * @return the full pathname to the requested library, or null
+   * @see Runtime#loadLibrary()
+   * @since 1.2
+   */
+  protected String findLibrary(String name)
+  {
+    return null;
+  }
+
+  /**
+   * Set the default assertion status for classes loaded by this classloader,
+   * used unless overridden by a package or class request.
+   *
+   * @param enabled true to set the default to enabled
+   * @see #setClassAssertionStatus(String, boolean)
+   * @see #setPackageAssertionStatus(String, boolean)
+   * @see #clearAssertionStatus()
+   * @since 1.4
+   */
+  public void setDefaultAssertionStatus(boolean enabled)
+  {
+    defaultAssertionStatus = enabled;
+  }
+  
+  /**
+   * Set the default assertion status for packages, used unless overridden
+   * by a class request. This default also covers subpackages, unless they
+   * are also specified. The unnamed package should use null for the name.
+   *
+   * @param name the package (and subpackages) to affect
+   * @param enabled true to set the default to enabled
+   * @see #setDefaultAssertionStatus(String, boolean)
+   * @see #setClassAssertionStatus(String, boolean)
+   * @see #clearAssertionStatus()
+   * @since 1.4
+   */
+  public synchronized void setPackageAssertionStatus(String name,
+                                                     boolean enabled)
+  {
+    if (packageAssertionStatus == null)
+      packageAssertionStatus
+        = new HashMap(StaticData.systemPackageAssertionStatus);
+    packageAssertionStatus.put(name, Boolean.valueOf(enabled));
+  }
+  
+  /**
+   * Set the default assertion status for a class. This only affects the
+   * status of top-level classes, any other string is harmless.
+   *
+   * @param name the class to affect
+   * @param enabled true to set the default to enabled
+   * @throws NullPointerException if name is null
+   * @see #setDefaultAssertionStatus(String, boolean)
+   * @see #setPackageAssertionStatus(String, boolean)
+   * @see #clearAssertionStatus()
+   * @since 1.4
+   */
+  public synchronized void setClassAssertionStatus(String name,
+                                                   boolean enabled)
+  {
+    if (classAssertionStatus == null)
+      classAssertionStatus = 
+        new HashMap(StaticData.systemClassAssertionStatus);
+    // The toString() hack catches null, as required.
+    classAssertionStatus.put(name.toString(), Boolean.valueOf(enabled));
+  }
+  
+  /**
+   * Resets the default assertion status of this classloader, its packages
+   * and classes, all to false. This allows overriding defaults inherited
+   * from the command line.
+   *
+   * @see #setDefaultAssertionStatus(boolean)
+   * @see #setClassAssertionStatus(String, boolean)
+   * @see #setPackageAssertionStatus(String, boolean)
+   * @since 1.4
+   */
+  public synchronized void clearAssertionStatus()
+  {
+    defaultAssertionStatus = false;
+    packageAssertionStatus = new HashMap();
+    classAssertionStatus = new HashMap();
+  }
+
+  /**
+   * Return true if this loader is either the specified class loader
+   * or an ancestor thereof.
+   * @param loader the class loader to check
+   */
+  final boolean isAncestorOf(ClassLoader loader)
+  {
+    while (loader != null)
+      {
+	if (this == loader)
+	  return true;
+	loader = loader.parent;
+      }
+    return false;
+  }
+
+  private static URL[] getExtClassLoaderUrls()
+  {
+    String classpath = SystemProperties.getProperty("java.ext.dirs", "");
+    StringTokenizer tok = new StringTokenizer(classpath, File.pathSeparator);
+    ArrayList list = new ArrayList();
+    while (tok.hasMoreTokens())
+      {
+	try
+	  {
+	    File f = new File(tok.nextToken());
+	    File[] files = f.listFiles();
+	    for (int i = 0; i < files.length; i++)
+	      {
+		list.add(files[i].toURL());
+	      }
+	  }
+	catch(Exception x)
+	  {
+	  }
+      }
+    URL[] urls = new URL[list.size()];
+    list.toArray(urls);
+    return urls;
+  }
+
+  private static void addFileURL(ArrayList list, String file)
+  {
+    try
+      {
+	list.add(new File(file).toURL());
+      }
+    catch(java.net.MalformedURLException x)
+      {
+      }
+  }
+
+  private static URL[] getSystemClassLoaderUrls()
+  {
+    String classpath = SystemProperties.getProperty("java.class.path", ".");
+    StringTokenizer tok = new StringTokenizer(classpath, File.pathSeparator, true);
+    ArrayList list = new ArrayList();
+    while (tok.hasMoreTokens())
+      {
+	String s = tok.nextToken();
+	if (s.equals(File.pathSeparator))
+	    addFileURL(list, ".");
+	else
+	  {
+	    addFileURL(list, s);
+	    if (tok.hasMoreTokens())
+	      {
+		// Skip the separator.
+		tok.nextToken();
+		// If the classpath ended with a separator,
+		// append the current directory.
+		if (!tok.hasMoreTokens())
+		    addFileURL(list, ".");
+	      }
+	  }
+      }
+    URL[] urls = new URL[list.size()];
+    list.toArray(urls);
+    return urls;
+  }
+
+  static ClassLoader defaultGetSystemClassLoader()
+  {
+    ClassLoader extClassLoader =
+	new URLClassLoader(getExtClassLoaderUrls(), null);
+    ClassLoader systemClassLoader =
+	new URLClassLoader(getSystemClassLoaderUrls(), extClassLoader)
+	{
+	    protected synchronized Class loadClass(String name,
+		boolean resolve)
+		throws ClassNotFoundException
+	    {
+		SecurityManager sm = SecurityManager.current;
+		if (sm != null)
+		{
+		    int lastDot = name.lastIndexOf('.');
+		    if (lastDot != -1)
+			sm.checkPackageAccess(name.substring(0, lastDot));
+		}
+		return super.loadClass(name, resolve);
+	    }
+	};
+    String loader = SystemProperties.getProperty("java.system.class.loader", null);
+    if (loader == null)
+      {
+	return systemClassLoader;
+      }
+    try
+      {
+	Constructor c = Class.forName(loader, false, systemClassLoader)
+	    .getConstructor(new Class[] { ClassLoader.class });
+	return (ClassLoader)c.newInstance(new Object[] { systemClassLoader });
+      }
+    catch (Exception e)
+      {
+	System.err.println("Requested system classloader " + loader + " failed.");
+	throw (Error)
+	    new Error("Requested system classloader " + loader + " failed.")
+		.initCause(e);
+      }
+  }
 }
Index: libraries/javalib/java/lang/NativeLibrary.java
===================================================================
RCS file: libraries/javalib/java/lang/NativeLibrary.java
diff -N libraries/javalib/java/lang/NativeLibrary.java
--- libraries/javalib/java/lang/NativeLibrary.java	22 Mar 2004 11:24:46 -0000	1.5
+++ /dev/null	1 Jan 1970 00:00:00 -0000
@@ -1,86 +0,0 @@
-
-/*
- * Java core library component.
- *
- * Copyright (c) 1999
- *	Archie L. Cobbs.  All rights reserved.
- * Copyright (c) 1999
- *	Transvirtual Technologies, Inc.  All rights reserved.
- *
- * See the file "license.terms" for information on usage and redistribution
- * of this file.
- *
- * Author: Archie L. Cobbs <archie@whistle.com>
- */
-
-package java.lang;
-
-import java.io.File;
-import java.io.FileNotFoundException;
-import java.util.StringTokenizer;
-
-/**
- * This class represents a linked in native library.
- */
-class NativeLibrary {
-
-	private int libIndex = -1;	// handle for native library
-
-	NativeLibrary(String path, ClassLoader loader)
-		throws FileNotFoundException {
-		if (loader == null)
-		{
-			// possible exception here
-			/*
-			 * No need to save the index since the root class
-			 * loader isn't going away so we won't ever unlink
-			 * the library.
-			 */
-			linkLibrary(path);
-		}
-		else
-		{
-			// possible exception here
-			libIndex = linkLibrary(path);
-			loader.addNativeLibrary(this);
-		}
-	}
-
-	static String[] getLibraryNames(String libname) {
-		final String libPath = 
-			System.getProperty("kaffe.library.path")
-			+ File.pathSeparatorChar
-			+ System.getProperty("java.library.path");
-
-		StringTokenizer t = new StringTokenizer(libPath,
-		    new String(new char[] { File.pathSeparatorChar }));
-		String[] dirs = new String[t.countTokens()];
-		for (int i = 0; i < dirs.length; i++) {
-			dirs[i] = t.nextToken() + File.separator
-			   + System.mapLibraryName(libname);
-		}
-		return dirs;
-	}
-
-	protected void finalize() throws Throwable {
-		if (libIndex != -1)
-			unlinkLibrary(libIndex);
-		super.finalize();
-	}
-
-	private static native int linkLibrary(String path)
-		throws FileNotFoundException;
-	
-	private static native void unlinkLibrary(int index);
-
-	/**
-	 * @return The configured native library prefix (e.g. lib)
-	 */
-	static native String getLibPrefix();
-
-	/**
-	 * @return The configured native library suffix (e.g. .so)
-	 */
-	static native String getLibSuffix();
-}
-
Index: libraries/javalib/java/lang/Runtime.java
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/java/lang/Runtime.java,v
retrieving revision 1.30
diff -u -p -r1.30 Runtime.java
--- libraries/javalib/java/lang/Runtime.java	30 Jul 2004 11:52:27 -0000	1.30
+++ libraries/javalib/java/lang/Runtime.java	19 Dec 2004 14:26:34 -0000
@@ -148,7 +148,24 @@ public void loadLibrary(String libname) 
 		    ThreadStack.getCallersClassLoader(false));
 }
 
-void loadLibrary(String libname, ClassLoader loader) {
+   /**
+    * Tries to load and link a native library.
+    *
+    * @return true if successfull
+    * @throws UnsatisfiedLinkError if the library was found but could not be linked.
+    */
+   private static native boolean linkLibrary(String name, ClassLoader loader)
+      throws UnsatisfiedLinkError;
+
+   private static native String getLibPrefix ();
+   private static native String getLibSuffix ();
+
+   static String mapLibraryName (String name)
+   {
+      return getLibPrefix() + name + getLibSuffix();
+   }
+
+void loadLibrary(String libname, ClassLoader loader) throws UnsatisfiedLinkError {
 	SecurityManager sm = System.getSecurityManager();
 	if (sm != null)
 		sm.checkLink(libname);
@@ -156,20 +173,24 @@ void loadLibrary(String libname, ClassLo
 	String filename;
 	String[] names;
 
-	if (loader != null && (filename = loader.findLibrary(libname)) != null)
-		names = new String[] { filename };
-	else
-		names = NativeLibrary.getLibraryNames(libname);
-	for (int i = 0; i < names.length; i++) {
-		try {
-			new NativeLibrary(names[i], loader);
+	if (loader != null && (filename = loader.findLibrary(libname)) != null) {
+		if (linkLibrary(filename, loader))
 			return;
-		} catch (FileNotFoundException e) {
-		} catch (UnsatisfiedLinkError e) {
-			errmsg = e.getMessage();
+	} else {
+		String libPath = System.getProperty("kaffe.library.path")
+				 + File.pathSeparatorChar
+				 + System.getProperty("java.library.path");
+
+		StringTokenizer t = new StringTokenizer(libPath,
+					new String(new char[] { File.pathSeparatorChar }));
+		while (t.hasMoreTokens ()) {
+			if (linkLibrary(t.nextToken() + File.separator + mapLibraryName(libname),
+					loader))
+				return;
 		}
 	}
-	throw new UnsatisfiedLinkError(errmsg + "\nAdding its directory to LD_LIBRARY_PATH may help.");
+	throw new UnsatisfiedLinkError("Could not find library '" + libname +
+				       "'\nAdding its directory to LD_LIBRARY_PATH may help.");
 }
 
 public void load(String filename) {
@@ -180,12 +201,9 @@ void load(String filename, ClassLoader l
 	SecurityManager sm = System.getSecurityManager();
 	if (sm != null)
 		sm.checkLink(filename);
-	try {
-		new NativeLibrary(filename, loader);
-	}
-	catch (FileNotFoundException e) {
+
+	if (!linkLibrary(filename, loader))
 		throw new UnsatisfiedLinkError(filename + ": not found");
-	}
 }
 
 native public long maxMemory();
Index: libraries/javalib/java/lang/System.java
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/java/lang/System.java,v
retrieving revision 1.42
diff -u -p -r1.42 System.java
--- libraries/javalib/java/lang/System.java	10 Dec 2004 21:22:54 -0000	1.42
+++ libraries/javalib/java/lang/System.java	19 Dec 2004 14:26:34 -0000
@@ -40,7 +40,7 @@ public static native void debugE(Throwab
 ****/
 static {
 	// Initialise the I/O
-	if (getProperty("kaffe.embedded", "false").equals("false")) {
+	if (SystemProperties.getProperty("kaffe.embedded", "false").equals("false")) {
 		in = new BufferedInputStream(new FileInputStream(FileDescriptor.in), 128);
 		out = new PrintStream(new BufferedOutputStream(new FileOutputStream(FileDescriptor.out), 128), true);
 		err = new PrintStream(new BufferedOutputStream(new FileOutputStream(FileDescriptor.err), 128), true);
@@ -134,7 +134,7 @@ public static void loadLibrary(String li
 }
 
 public static String mapLibraryName(String fn) {
-	return NativeLibrary.getLibPrefix() + fn + NativeLibrary.getLibSuffix();
+	return Runtime.getRuntime().mapLibraryName(fn);
 }
 
 public static void runFinalization() {
Index: libraries/javalib/java/lang/VMClassLoader.java
===================================================================
RCS file: libraries/javalib/java/lang/VMClassLoader.java
diff -N libraries/javalib/java/lang/VMClassLoader.java
--- /dev/null	1 Jan 1970 00:00:00 -0000
+++ libraries/javalib/java/lang/VMClassLoader.java	19 Dec 2004 14:26:34 -0000
@@ -0,0 +1,299 @@
+/* VMClassLoader.java -- Reference implementation of native interface
+   required by ClassLoader
+   Copyright (C) 1998, 2001, 2002, 2004 Free Software Foundation
+
+This file is part of GNU Classpath.
+
+GNU Classpath is free software; you can redistribute it and/or modify
+it under the terms of the GNU General Public License as published by
+the Free Software Foundation; either version 2, or (at your option)
+any later version.
+
+GNU Classpath is distributed in the hope that it will be useful, but
+WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with GNU Classpath; see the file COPYING.  If not, write to the
+Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+02111-1307 USA.
+
+Linking this library statically or dynamically with other modules is
+making a combined work based on this library.  Thus, the terms and
+conditions of the GNU General Public License cover the whole
+combination.
+
+As a special exception, the copyright holders of this library give you
+permission to link this library with independent modules to produce an
+executable, regardless of the license terms of these independent
+modules, and to copy and distribute the resulting executable under
+terms of your choice, provided that you also meet, for each linked
+independent module, the terms and conditions of the license of that
+module.  An independent module is a module which is not derived from
+or based on this library.  If you modify this library, you may extend
+this exception to your version of the library, but you are not
+obligated to do so.  If you do not wish to do so, delete this
+exception statement from your version. */
+
+package java.lang;
+
+import java.security.ProtectionDomain;
+import java.net.URL;
+import java.net.MalformedURLException;
+import java.io.IOException;
+import java.io.File;
+import java.util.Enumeration;
+import java.util.Map;
+import java.util.HashMap;
+import java.util.StringTokenizer;
+import java.util.Vector;
+import java.lang.reflect.Constructor;
+
+import gnu.java.util.EmptyEnumeration;
+
+import kaffe.lang.PrimordialClassLoader;
+import kaffe.lang.AppClassLoader;
+
+/**
+ * java.lang.VMClassLoader is a package-private helper for VMs to implement
+ * on behalf of java.lang.ClassLoader.
+ *
+ * @author John Keiser
+ * @author Mark Wielaard <mark@klomp.org>
+ * @author Eric Blake <ebb9@email.byu.edu>
+ */
+final class VMClassLoader
+{
+  /**
+   * Helper to define a class using a string of bytes. This assumes that
+   * the security checks have already been performed, if necessary.
+   *
+   * <strong>For backward compatibility, this just ignores the protection
+   * domain; that is the wrong behavior, and you should directly implement
+   * this method natively if you can.</strong>
+   *
+   * Implementations of this method are advised to consider the
+   * situation where user code modifies the byte array after it has
+   * been passed to defineClass.  This can be handled by making a
+   * private copy of the array, or arranging to only read any given
+   * byte a single time.
+   *
+   * @param name the name to give the class, or null if unknown
+   * @param data the data representing the classfile, in classfile format
+   * @param offset the offset into the data where the classfile starts
+   * @param len the length of the classfile data in the array
+   * @param pd the protection domain
+   * @return the class that was defined
+   * @throws ClassFormatError if data is not in proper classfile format
+   */
+  static final native Class defineClass(ClassLoader cl, String name,
+					byte[] data, int offset, int len,
+					ProtectionDomain pd)
+    throws ClassFormatError;
+
+  /**
+   * Helper to resolve all references to other classes from this class.
+   *
+   * @param c the class to resolve
+   */
+  static final native void resolveClass(Class c);
+
+  /**
+   * Helper to load a class from the bootstrap class loader.
+   *
+   * XXX - Not implemented; this requires native help.
+   *
+   * @param name the class name to load
+   * @param resolve whether to resolve it
+   * @return the class, loaded by the bootstrap classloader or null
+   * if the class wasn't found. Returning null is equivalent to throwing
+   * a ClassNotFoundException (but a possible performance optimization).
+   */
+  static Class loadClass(String name, boolean resolve)
+    throws ClassNotFoundException
+  {
+    return PrimordialClassLoader.getSingleton().loadClass(name, resolve);
+  }
+
+  /**
+   * Helper to load a resource from the bootstrap class loader.
+   *
+   * @param name the resource to find
+   * @return the URL to the resource
+   */
+  static URL getResource(String name)
+  {
+    return PrimordialClassLoader.getSingleton().getResource(name);
+  }
+
+  /**
+   * Helper to get a list of resources from the bootstrap class loader.
+   *
+   * @param name the resource to find
+   * @return an enumeration of resources
+   * @throws IOException if one occurs
+   */
+  static Enumeration getResources(String name) throws IOException
+  {
+    /* we can't override getResources in PrimordialClassLoader as it's final.
+     * So this should not delegate to the PrimordialLoader as it will also be
+     * called for the PrimordialClassLoader singleton itself.
+     */
+    return EmptyEnumeration.getInstance();
+  }
+
+  /**
+   * Helper to get a package from the bootstrap class loader.  The default
+   * implementation of returning null may be adequate, or you may decide
+   * that this needs some native help.
+   *
+   * @param name the name to find
+   * @return the named package, if it exists
+   */
+  static Package getPackage(String name)
+  {
+    return PrimordialClassLoader.getSingleton().getPackage(name);
+  }
+
+  /**
+   * Helper to get all packages from the bootstrap class loader.  The default
+   * implementation of returning an empty array may be adequate, or you may
+   * decide that this needs some native help.
+   *
+   * @return all named packages, if any exist
+   */
+  static Package[] getPackages()
+  {
+    return PrimordialClassLoader.getSingleton().getPackages();
+  }
+
+  /**
+   * Helper for java.lang.Integer, Byte, etc to get the TYPE class
+   * at initialization time. The type code is one of the chars that
+   * represents the primitive type as in JNI.
+   *
+   * <ul>
+   * <li>'Z' - boolean</li>
+   * <li>'B' - byte</li>
+   * <li>'C' - char</li>
+   * <li>'D' - double</li>
+   * <li>'F' - float</li>
+   * <li>'I' - int</li>
+   * <li>'J' - long</li>
+   * <li>'S' - short</li>
+   * <li>'V' - void</li>
+   * </ul>
+   *
+   * Note that this is currently a java version that converts the type code
+   * to a string and calls the native <code>getPrimitiveClass(String)</code>
+   * method for backwards compatibility with VMs that used old versions of
+   * GNU Classpath. Please replace this method with a native method
+   * <code>final static native Class getPrimitiveClass(char type);</code>
+   * if your VM supports it. <strong>The java version of this method and
+   * the String version of this method will disappear in a future version
+   * of GNU Classpath</strong>.
+   *
+   * @param type the primitive type
+   * @return a "bogus" class representing the primitive type
+   */
+  static final Class getPrimitiveClass(char type)
+  {
+    String t;
+    switch (type)
+      {
+      case 'Z':
+        t = "boolean";
+        break;
+      case 'B':
+        t = "byte";
+        break;
+      case 'C':
+        t = "char";
+        break;
+      case 'D':
+        t = "double";
+        break;
+      case 'F':
+        t = "float";
+        break;
+      case 'I':
+        t = "int";
+        break;
+      case 'J':
+        t = "long";
+        break;
+      case 'S':
+        t = "short";
+        break;
+      case 'V':
+        t = "void";
+        break;
+      default:
+        throw new NoClassDefFoundError("Invalid type specifier: " + type);
+      }
+    return getPrimitiveClass(t);
+  }
+
+  /**
+   * Old version of the interface, added here for backwards compatibility.
+   * Called by the java version of getPrimitiveClass(char) when no native
+   * version of that method is available.
+   * <strong>This method will be removed in a future version of GNU
+   * Classpath</strong>.
+   * @param type the primitive type
+   * @return a "bogus" class representing the primitive type
+   */
+  static Class getPrimitiveClass(String type)
+  {
+    throw new InternalError ();
+  }
+
+  /**
+   * The system default for assertion status. This is used for all system
+   * classes (those with a null ClassLoader), as well as the initial value for
+   * every ClassLoader's default assertion status.
+   *
+   * XXX - Not implemented yet; this requires native help.
+   *
+   * @return the system-wide default assertion status
+   */
+  static final boolean defaultAssertionStatus()
+  {
+    return true;
+  }
+
+  /**
+   * The system default for package assertion status. This is used for all
+   * ClassLoader's packageAssertionStatus defaults. It must be a map of
+   * package names to Boolean.TRUE or Boolean.FALSE, with the unnamed package
+   * represented as a null key.
+   *
+   * XXX - Not implemented yet; this requires native help.
+   *
+   * @return a (read-only) map for the default packageAssertionStatus
+   */
+  static final Map packageAssertionStatus()
+  {
+    return new HashMap();
+  }
+
+  /**
+   * The system default for class assertion status. This is used for all
+   * ClassLoader's classAssertionStatus defaults. It must be a map of
+   * class names to Boolean.TRUE or Boolean.FALSE
+   *
+   * XXX - Not implemented yet; this requires native help.
+   *
+   * @return a (read-only) map for the default classAssertionStatus
+   */
+  static final Map classAssertionStatus()
+  {
+    return new HashMap();
+  }
+
+  static ClassLoader getSystemClassLoader ()
+  {
+    return AppClassLoader.getSingleton();
+  }
+}
Index: libraries/javalib/kaffe/lang/AppClassLoader.java
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/kaffe/lang/AppClassLoader.java,v
retrieving revision 1.6
diff -u -p -r1.6 AppClassLoader.java
--- libraries/javalib/kaffe/lang/AppClassLoader.java	22 Mar 2004 11:25:32 -0000	1.6
+++ libraries/javalib/kaffe/lang/AppClassLoader.java	19 Dec 2004 14:26:34 -0000
@@ -210,9 +210,9 @@ public void addSource(String path) {
 	if (!f.exists()) {
 		return;
 	}
-	
+
 	Source s = null;
-	
+
 	try {
 		if (f.isDirectory()) {
 			s = new DirSource (f.getAbsoluteFile ());
@@ -234,7 +234,7 @@ public void addSource(String path) {
 }
 	
 private AppClassLoader() {
-	super(new URL[0]);
+	super(new URL[0], PrimordialClassLoader.getSingleton());
 
 	StringTokenizer tok = new StringTokenizer (System.getProperty("java.class.path"), File.pathSeparator);
 
Index: libraries/javalib/kaffe/lang/PrimordialClassLoader.java
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/kaffe/lang/PrimordialClassLoader.java,v
retrieving revision 1.6
diff -u -p -r1.6 PrimordialClassLoader.java
--- libraries/javalib/kaffe/lang/PrimordialClassLoader.java	31 May 2004 21:15:33 -0000	1.6
+++ libraries/javalib/kaffe/lang/PrimordialClassLoader.java	19 Dec 2004 14:26:34 -0000
@@ -56,7 +56,8 @@ public Class loadClass(String name, bool
 {
 	Class retval;
 
-	if( name.startsWith("kaffe.lang.") )
+	if( name.startsWith("kaffe.lang.") ||
+            name.startsWith("gnu.classpath.") )
 	{
 		throw new ClassNotFoundException(name);
 	}
@@ -165,7 +166,7 @@ protected void findResources(Vector v, S
 	}
 }
 
-public URL findResource (String name) {
+public URL getResource (String name) {
 	try {
 		Vector v = new Vector ();
 
Index: libraries/javalib/kaffe/lang/ThreadStack.java
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/javalib/kaffe/lang/ThreadStack.java,v
retrieving revision 1.3
diff -u -p -r1.3 ThreadStack.java
--- libraries/javalib/kaffe/lang/ThreadStack.java	31 Oct 2004 14:35:31 -0000	1.3
+++ libraries/javalib/kaffe/lang/ThreadStack.java	19 Dec 2004 14:26:36 -0000
@@ -84,7 +84,7 @@ public class ThreadStack
 		 * we return the AppClassLoader in this case.
 		 */
 		if (frameIdx == classStack.length) {
-			return AppClassLoader.getSingleton();
+			return ClassLoader.getSystemClassLoader();
 		}
 
 		/*
@@ -97,7 +97,7 @@ public class ThreadStack
 		 * this should not happen ....
 		 */
 		if (frameIdx==classStack.length) {
-			return javaPrimordial ? PrimordialClassLoader.getSingleton() : null;
+			return null;
 		}
 
 		/*
@@ -112,7 +112,7 @@ public class ThreadStack
 
 		ClassLoader ret = classStack[frameIdx].getClassLoader();
 
-		return (ret==null && javaPrimordial) ? PrimordialClassLoader.getSingleton() : ret; 
+		return (ret==null && javaPrimordial) ? PrimordialClassLoader.getSingleton() : ret;
 	}
 
 	/**
Index: libraries/javalib/gnu/java/util/DoubleEnumeration.java
===================================================================
RCS file: libraries/javalib/gnu/java/util/DoubleEnumeration.java
diff -N libraries/javalib/gnu/java/util/DoubleEnumeration.java
--- /dev/null	1 Jan 1970 00:00:00 -0000
+++ libraries/javalib/gnu/java/util/DoubleEnumeration.java	19 Dec 2004 14:26:36 -0000
@@ -0,0 +1,138 @@
+/* gnu.java.util.DoubleEnumeration
+   Copyright (C) 1998, 1999, 2001 Free Software Foundation, Inc.
+
+This file is part of GNU Classpath.
+
+GNU Classpath is free software; you can redistribute it and/or modify
+it under the terms of the GNU General Public License as published by
+the Free Software Foundation; either version 2, or (at your option)
+any later version.
+
+GNU Classpath is distributed in the hope that it will be useful, but
+WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with GNU Classpath; see the file COPYING.  If not, write to the
+Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+02111-1307 USA.
+
+Linking this library statically or dynamically with other modules is
+making a combined work based on this library.  Thus, the terms and
+conditions of the GNU General Public License cover the whole
+combination.
+
+As a special exception, the copyright holders of this library give you
+permission to link this library with independent modules to produce an
+executable, regardless of the license terms of these independent
+modules, and to copy and distribute the resulting executable under
+terms of your choice, provided that you also meet, for each linked
+independent module, the terms and conditions of the license of that
+module.  An independent module is a module which is not derived from
+or based on this library.  If you modify this library, you may extend
+this exception to your version of the library, but you are not
+obligated to do so.  If you do not wish to do so, delete this
+exception statement from your version. */
+
+package gnu.java.util;
+
+import java.util.Enumeration;
+import java.util.NoSuchElementException;
+
+
+/**
+ * This is a helper class that combines two Enumerations.
+ * It returns the elements of the first Enumeration until it has
+ * no more elements and then returns the elements of the second
+ * Enumeration.<br>
+ * 
+ * In the default case:
+ * <pre>
+ * doubleEnum = new DoubleEnumeration(enum1, enum2);
+ * while (doubleEnum.hasMoreElements()) {
+ *    Object o = doubleEnum.nextElement();
+ *    do_something(o);
+ * }
+ * </pre>
+ * it calls hasMoreElements of the Enumerations as few times as
+ * possible.
+ * The references to the Enumerations are cleared as soon as they have no
+ * more elements to help garbage collecting.
+ *
+ * @author Jochen Hoenicke
+ * @author Mark Wielaard (mark@klomp.org)
+ */
+public class DoubleEnumeration implements Enumeration
+{
+  /**
+   * This is true as long as one of the enumerations has more
+   * elements.  
+   * Only valid when hasChecked is true.
+   * Set in <code>hasMoreElements()</code>
+   */
+  private boolean hasMore;
+  /**
+   * This is true, if it is sure that hasMore indicates wether there are
+   * more elements.
+   * Set to true in <code>hasMoreElements()</code>.
+   * Set to false in <code>getNextElement()</code>.
+   */
+  private boolean hasChecked;
+  /**
+   * The first enumeration.
+   */
+  private Enumeration e1;
+  /**
+   * The second enumeration.
+   */
+  private Enumeration e2;
+
+  /**
+   * Creates a new Enumeration combining the given two enumerations.
+   * The enumerations mustn't be accessed by other classes.
+   */
+  public DoubleEnumeration(Enumeration e1, Enumeration e2)
+  {
+    this.e1 = e1;
+    this.e2 = e2;
+    hasChecked = false;
+  }
+
+  /**
+   * Returns true, if at least one of the two enumerations has more
+   * elements.
+   */
+  public boolean hasMoreElements()
+  {
+    if (hasChecked)
+      return hasMore;
+
+    hasMore = (e1 != null && e1.hasMoreElements());
+
+    if (!hasMore) {
+      e1 = e2;
+      e2 = null;
+      hasMore = (e1 != null && e1.hasMoreElements());
+    }
+
+    hasChecked = true;
+    return hasMore;
+  }
+
+  /**
+   * Returns the next element.  This returns the next element of the
+   * first enumeration, if it has more elements, otherwise the next
+   * element of the second enumeration. If both enumeration don't have
+   * any elements it throws a <code>NoSuchElementException</code>.
+   */
+  public Object nextElement()
+  {
+    if (!hasMoreElements())
+      throw new NoSuchElementException();
+    else {
+      hasChecked = false;
+      return e1.nextElement();
+    }
+  }
+}
Index: libraries/clib/native/Class.c
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/clib/native/Class.c,v
retrieving revision 1.75
diff -u -p -r1.75 Class.c
--- libraries/clib/native/Class.c	5 Dec 2004 22:43:29 -0000	1.75
+++ libraries/clib/native/Class.c	19 Dec 2004 14:26:36 -0000
@@ -289,7 +289,13 @@ java_lang_Class_getSigners0(struct Hjava
 void
 java_lang_Class_setSigners(struct Hjava_lang_Class* this, HArrayOfObject* sigs)
 {
-	this->signers = sigs;
+	unhand(this)->signers = sigs;
+}
+
+struct Hjava_lang_Object*
+java_lang_Class_getProtectionDomain (struct Hjava_lang_Class* this)
+{
+	return unhand(this)->protectionDomain;
 }
 
 /*
Index: libraries/clib/native/ClassLoader.c
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/clib/native/ClassLoader.c,v
retrieving revision 1.39
diff -u -p -r1.39 ClassLoader.c
--- libraries/clib/native/ClassLoader.c	21 May 2004 15:19:24 -0000	1.39
+++ libraries/clib/native/ClassLoader.c	19 Dec 2004 14:26:36 -0000
@@ -31,7 +31,8 @@
  * Translate an array of bytes into a class.
  */
 struct Hjava_lang_Class*
-java_lang_ClassLoader_defineClass0(struct Hjava_lang_ClassLoader* this, struct Hjava_lang_String* name, HArrayOfByte* data, jint offset, jint length)
+java_lang_VMClassLoader_defineClass(struct Hjava_lang_ClassLoader* this, struct Hjava_lang_String* name,
+				    HArrayOfByte* data, jint offset, jint length, struct Hjava_lang_Object* protectionDomain)
 {
 	Hjava_lang_Class *clazz, *dup_clazz;
 	classFile hand;
@@ -67,6 +68,9 @@ java_lang_ClassLoader_defineClass0(struc
 		throwError(&info);
 	}
 
+	/* set protection domain of new class */
+	unhand(clazz)->protectionDomain = protectionDomain;
+
 	/*
 	 * See if an entry for that name and class loader already exists
 	 * create one if not.
@@ -116,9 +120,10 @@ java_lang_ClassLoader_defineClass0(struc
  * Resolve classes reference by this class.
  */
 void
-java_lang_ClassLoader_resolveClass0(struct Hjava_lang_ClassLoader* this UNUSED, struct Hjava_lang_Class* class)
+java_lang_VMClassLoader_resolveClass(struct Hjava_lang_Class* class)
 {
 	errorInfo info;
+
 	if (processClass(class, CSTATE_COMPLETE, &info) == false) {
 		throwError(&info);
 	}
Index: libraries/clib/native/Makefile.am
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/clib/native/Makefile.am,v
retrieving revision 1.37
diff -u -p -r1.37 Makefile.am
--- libraries/clib/native/Makefile.am	8 Dec 2004 07:22:12 -0000	1.37
+++ libraries/clib/native/Makefile.am	19 Dec 2004 14:26:36 -0000
@@ -29,7 +29,6 @@ LANG_SRCS = \
 		Float.c \
 		Math.c \
 		MemoryAdvice.c \
-		NativeLibrary.c \
 		Object.c \
 		PrimordialClassLoader.c \
 		Runtime.c \
Index: libraries/clib/native/Makefile.in
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/clib/native/Makefile.in,v
retrieving revision 1.163
diff -u -p -r1.163 Makefile.in
--- libraries/clib/native/Makefile.in	8 Dec 2004 07:22:12 -0000	1.163
+++ libraries/clib/native/Makefile.in	19 Dec 2004 14:26:36 -0000
@@ -95,7 +95,7 @@ am__objects_1 = libnative_la-ByteToCharD
 am__objects_2 = libnative_la-Class.lo libnative_la-ClassLoader.lo \
 	libnative_la-Double.lo libnative_la-Float.lo \
 	libnative_la-Math.lo libnative_la-MemoryAdvice.lo \
-	libnative_la-NativeLibrary.lo libnative_la-Object.lo \
+	libnative_la-Object.lo \
 	libnative_la-PrimordialClassLoader.lo libnative_la-Runtime.lo \
 	libnative_la-String.lo libnative_la-System.lo \
 	libnative_la-Thread.lo libnative_la-ThreadStack.lo \
@@ -412,7 +412,6 @@ LANG_SRCS = \
 		Float.c \
 		Math.c \
 		MemoryAdvice.c \
-		NativeLibrary.c \
 		Object.c \
 		PrimordialClassLoader.c \
 		Runtime.c \
@@ -548,7 +547,6 @@ distclean-compile:
 @AMDEP_TRUE@@am__include@ @am__quote@./$(DEPDIR)/libnative_la-Math.Plo@am__quote@
 @AMDEP_TRUE@@am__include@ @am__quote@./$(DEPDIR)/libnative_la-MemoryAdvice.Plo@am__quote@
 @AMDEP_TRUE@@am__include@ @am__quote@./$(DEPDIR)/libnative_la-Method.Plo@am__quote@
-@AMDEP_TRUE@@am__include@ @am__quote@./$(DEPDIR)/libnative_la-NativeLibrary.Plo@am__quote@
 @AMDEP_TRUE@@am__include@ @am__quote@./$(DEPDIR)/libnative_la-Object.Plo@am__quote@
 @AMDEP_TRUE@@am__include@ @am__quote@./$(DEPDIR)/libnative_la-PrimordialClassLoader.Plo@am__quote@
 @AMDEP_TRUE@@am__include@ @am__quote@./$(DEPDIR)/libnative_la-Runtime.Plo@am__quote@
@@ -653,13 +651,6 @@ libnative_la-MemoryAdvice.lo: MemoryAdvi
 @AMDEP_TRUE@@am__fastdepCC_FALSE@	source='MemoryAdvice.c' object='libnative_la-MemoryAdvice.lo' libtool=yes @AMDEPBACKSLASH@
 @AMDEP_TRUE@@am__fastdepCC_FALSE@	DEPDIR=$(DEPDIR) $(CCDEPMODE) $(depcomp) @AMDEPBACKSLASH@
 @am__fastdepCC_FALSE@	$(LIBTOOL) --tag=CC --mode=compile $(CC) $(DEFS) $(DEFAULT_INCLUDES) $(INCLUDES) $(AM_CPPFLAGS) $(CPPFLAGS) $(libnative_la_CFLAGS) $(CFLAGS) -c -o libnative_la-MemoryAdvice.lo `test -f 'MemoryAdvice.c' || echo '$(srcdir)/'`MemoryAdvice.c
-
-libnative_la-NativeLibrary.lo: NativeLibrary.c
-@am__fastdepCC_TRUE@	if $(LIBTOOL) --tag=CC --mode=compile $(CC) $(DEFS) $(DEFAULT_INCLUDES) $(INCLUDES) $(AM_CPPFLAGS) $(CPPFLAGS) $(libnative_la_CFLAGS) $(CFLAGS) -MT libnative_la-NativeLibrary.lo -MD -MP -MF "$(DEPDIR)/libnative_la-NativeLibrary.Tpo" -c -o libnative_la-NativeLibrary.lo `test -f 'NativeLibrary.c' || echo '$(srcdir)/'`NativeLibrary.c; \
-@am__fastdepCC_TRUE@	then mv -f "$(DEPDIR)/libnative_la-NativeLibrary.Tpo" "$(DEPDIR)/libnative_la-NativeLibrary.Plo"; else rm -f "$(DEPDIR)/libnative_la-NativeLibrary.Tpo"; exit 1; fi
-@AMDEP_TRUE@@am__fastdepCC_FALSE@	source='NativeLibrary.c' object='libnative_la-NativeLibrary.lo' libtool=yes @AMDEPBACKSLASH@
-@AMDEP_TRUE@@am__fastdepCC_FALSE@	DEPDIR=$(DEPDIR) $(CCDEPMODE) $(depcomp) @AMDEPBACKSLASH@
-@am__fastdepCC_FALSE@	$(LIBTOOL) --tag=CC --mode=compile $(CC) $(DEFS) $(DEFAULT_INCLUDES) $(INCLUDES) $(AM_CPPFLAGS) $(CPPFLAGS) $(libnative_la_CFLAGS) $(CFLAGS) -c -o libnative_la-NativeLibrary.lo `test -f 'NativeLibrary.c' || echo '$(srcdir)/'`NativeLibrary.c
 
 libnative_la-Object.lo: Object.c
 @am__fastdepCC_TRUE@	if $(LIBTOOL) --tag=CC --mode=compile $(CC) $(DEFS) $(DEFAULT_INCLUDES) $(INCLUDES) $(AM_CPPFLAGS) $(CPPFLAGS) $(libnative_la_CFLAGS) $(CFLAGS) -MT libnative_la-Object.lo -MD -MP -MF "$(DEPDIR)/libnative_la-Object.Tpo" -c -o libnative_la-Object.lo `test -f 'Object.c' || echo '$(srcdir)/'`Object.c; \
Index: libraries/clib/native/NativeLibrary.c
===================================================================
RCS file: libraries/clib/native/NativeLibrary.c
diff -N libraries/clib/native/NativeLibrary.c
--- libraries/clib/native/NativeLibrary.c	11 Dec 2004 07:00:52 -0000	1.5
+++ /dev/null	1 Jan 1970 00:00:00 -0000
@@ -1,95 +0,0 @@
-
-/*
- * java.lang.NativeLibrary.c
- *
- * Copyright (c) 2000
- *	Archie L. Cobbs.  All rights reserved.
- * Copyright (c) 2000
- *	Transvirtual Technologies, Inc.  All rights reserved.
- *
- * See the file "license.terms" for information on usage and redistribution 
- * of this file. 
- */
-
-#include "config.h"
-#include "config-std.h"
-#include "config-io.h"
-#include "config-mem.h"
-#include "defs.h"
-#include "gtypes.h"
-#include "object.h"
-#include "constants.h"
-#include "support.h"
-#include "external.h"
-#include "java_lang_String.h"
-#include "java_lang_NativeLibrary.h"
-#include <native.h>
-
-#ifndef LIBRARYPREFIX
-#define	LIBRARYPREFIX	"lib"
-#endif
-
-#ifndef LIBRARYSUFFIX
-#define LIBRARYSUFFIX	""
-#endif
-
-/*
- * Return prefix to prepend to library names
- */
-struct Hjava_lang_String*
-java_lang_NativeLibrary_getLibPrefix()
-{
-	return checkPtr(stringC2Java(LIBRARYPREFIX));
-}
-
-/*
- * Return suffix to append to library names
- */
-struct Hjava_lang_String*
-java_lang_NativeLibrary_getLibSuffix()
-{
-	return checkPtr(stringC2Java(LIBRARYSUFFIX));
-}
-
-/*
- * Attempt to link in a shared library. Throws an UnsatisfiedLinkError
- * if the attempt fails.
- */
-int
-java_lang_NativeLibrary_linkLibrary(struct Hjava_lang_String *jpath)
-{
-	char path[MAXPATHLEN];
-	char errbuf[128];
-	errorInfo einfo;
-	int i = -1;
-
-	stringJava2CBuf(jpath, path, sizeof(path));
-	if ((i = loadNativeLibrary(path, errbuf, sizeof(errbuf))) < 0) {
- 		if( strstr(errbuf, "File not found") ) {
- 			postExceptionMessage(&einfo,
- 					     JAVA_IO(FileNotFoundException),
- 					     "%s",
- 					     path);
- 		} else {
-			postExceptionMessage(&einfo, 
-					     JAVA_LANG(UnsatisfiedLinkError),
-					     "%s",
-					     errbuf);
-			throwError(&einfo);
-		}
-	}
- 	if( i == -1 )
- 		throwError(&einfo);
-
-	return i;
-}
-
-/*
- * Unlink a shared library.
- */
-void
-java_lang_NativeLibrary_unlinkLibrary(int i)
-{
-	unloadNativeLibrary(i);
-}
-
Index: libraries/clib/native/Runtime.c
===================================================================
RCS file: /cvs/kaffe/kaffe/libraries/clib/native/Runtime.c,v
retrieving revision 1.28
diff -u -p -r1.28 Runtime.c
--- libraries/clib/native/Runtime.c	17 Dec 2004 03:38:50 -0000	1.28
+++ libraries/clib/native/Runtime.c	19 Dec 2004 14:26:36 -0000
@@ -17,11 +17,21 @@
 #include "files.h"
 #include "defs.h"
 #include "java_io_File.h"
+#include "java_lang_String.h"
 #include "java_lang_Runtime.h"
 #include "external.h"
 #include "gc.h"
 #include "support.h"
 #include "stringSupport.h"
+#include "external.h"
+
+#ifndef LIBRARYPREFIX
+#define	LIBRARYPREFIX	"lib"
+#endif
+
+#ifndef LIBRARYSUFFIX
+#define LIBRARYSUFFIX	""
+#endif
 
 extern jboolean runFinalizerOnExit;
 
@@ -130,4 +140,47 @@ void
 java_lang_Runtime_runFinalizersOnExit(jboolean on)
 {
 	runFinalizerOnExit = on;
+}
+
+
+/*
+ * Attempt to link in a shared library. Throws an UnsatisfiedLinkError
+ * if the attempt fails.
+ */
+jboolean
+java_lang_Runtime_linkLibrary(struct Hjava_lang_String *jpath, struct Hjava_lang_ClassLoader* loader)
+{
+	char path[MAXPATHLEN];
+	char errbuf[128];
+	errorInfo einfo;
+
+	stringJava2CBuf(jpath, path, sizeof(path));
+	if (loadNativeLibrary(path, loader, errbuf, sizeof(errbuf)) < 0) {
+ 		if( strstr(errbuf, ": not found") ) {
+ 			postExceptionMessage(&einfo,
+ 					     JAVA_IO(FileNotFoundException),
+ 					     "%s",
+ 					     path);
+			return false;
+ 		} else {
+			postExceptionMessage(&einfo, 
+					     JAVA_LANG(UnsatisfiedLinkError),
+					     "%s",
+					     errbuf);
+			throwError(&einfo);
+		}
+	}
+	return true;
+}
+
+struct Hjava_lang_String*
+java_lang_Runtime_getLibPrefix()
+{
+	return checkPtr(stringC2Java(LIBRARYPREFIX));
+}
+
+struct Hjava_lang_String*
+java_lang_Runtime_getLibSuffix()
+{
+	return checkPtr(stringC2Java(LIBRARYSUFFIX));
 }
Index: test/regression/KaffeInternal.java
===================================================================
RCS file: /cvs/kaffe/kaffe/test/regression/KaffeInternal.java,v
retrieving revision 1.1
diff -u -p -r1.1 KaffeInternal.java
--- test/regression/KaffeInternal.java	24 May 2003 20:09:04 -0000	1.1
+++ test/regression/KaffeInternal.java	19 Dec 2004 14:26:36 -0000
@@ -1,5 +1,6 @@
 
 import kaffe.lang.ThreadStack;
+import gnu.classpath.SystemProperties;
 
 /**
  * Test to make sure a user level class can't access kaffe internal classes.
@@ -16,9 +17,19 @@ class KaffeInternal
 	{
 	    System.out.println(th);
 	}
+
+        try
+        {
+            SystemProperties.getProperty("java.home");
+        }
+        catch(Throwable th)
+        {
+            System.out.println(th);
+        }
     }
 }
 
 /* Expected Output:
 java.lang.NoClassDefFoundError: kaffe/lang/ThreadStack
+java.lang.NoClassDefFoundError: gnu/classpath/SystemProperties
 */
Index: test/regression/Makefile.am
===================================================================
RCS file: /cvs/kaffe/kaffe/test/regression/Makefile.am,v
retrieving revision 1.90
diff -u -p -r1.90 Makefile.am
--- test/regression/Makefile.am	8 Dec 2004 21:16:51 -0000	1.90
+++ test/regression/Makefile.am	19 Dec 2004 14:26:36 -0000
@@ -164,6 +164,7 @@ TEST_REFLECTION = \
 
 ## tests for ClassLoader
 TEST_CLASS_LOADING = \
+	PrimordialLoaderTest.java \
 	SystemLoaderTest.java \
 	NoClassDefTest.java \
 	CLTest.java \
Index: test/regression/Makefile.in
===================================================================
RCS file: /cvs/kaffe/kaffe/test/regression/Makefile.in,v
retrieving revision 1.195
diff -u -p -r1.195 Makefile.in
--- test/regression/Makefile.in	8 Dec 2004 21:16:51 -0000	1.195
+++ test/regression/Makefile.in	19 Dec 2004 14:26:36 -0000
@@ -493,6 +493,7 @@ TEST_REFLECTION = \
 	DeleteFile.java
 
 TEST_CLASS_LOADING = \
+	PrimordialLoaderTest.java \
 	SystemLoaderTest.java \
 	NoClassDefTest.java \
 	CLTest.java \
Index: test/regression/PrimordialLoaderTest.java
===================================================================
RCS file: test/regression/PrimordialLoaderTest.java
diff -N test/regression/PrimordialLoaderTest.java
--- /dev/null	1 Jan 1970 00:00:00 -0000
+++ test/regression/PrimordialLoaderTest.java	19 Dec 2004 14:26:36 -0000
@@ -0,0 +1,38 @@
+
+/*
+ * Checks that our primordial loader can properly find resources. 
+ */
+
+import java.net.URL;
+import java.net.JarURLConnection;
+
+public class PrimordialLoaderTest
+{
+
+	public static void main (String[] args) throws java.io.IOException
+	{
+		/* get a resource that's somewhere in the bootclasspath */
+		URL url = ClassLoader.getSystemResource("java/lang/Object.class");
+ 
+		System.out.println (url.getProtocol()+" "+((JarURLConnection)url.openConnection()).getEntryName());
+
+		java.util.Enumeration enum = ClassLoader.getSystemResources("java/lang/Object.class");
+
+		while (enum.hasMoreElements())
+		{
+			url = (URL)enum.nextElement();
+
+			System.out.println (url.getProtocol()+" "+((JarURLConnection)url.openConnection()).getEntryName());
+		}
+
+		System.out.println (ClassLoader.getSystemResourceAsStream("java/lang/Object.class").getClass());
+	}
+
+}
+
+
+/* Expected Output:
+jar java/lang/Object.class
+jar java/lang/Object.class
+class java.util.jar.JarFile$EntryInputStream
+*/
Index: test/regression/ProcessClassInst.java
===================================================================
RCS file: /cvs/kaffe/kaffe/test/regression/ProcessClassInst.java,v
retrieving revision 1.11
diff -u -p -r1.11 ProcessClassInst.java
--- test/regression/ProcessClassInst.java	5 Dec 2004 22:43:32 -0000	1.11
+++ test/regression/ProcessClassInst.java	19 Dec 2004 14:26:36 -0000
@@ -104,6 +104,6 @@ tan java.lang.NoClassDefFoundError: Base
 java.lang.NoClassDefFoundError: Base
    at java.lang.Class.getConstructor0 (Class.java)
    at java.lang.Class.getDeclaredConstructor (Class.java:174)
-   at java.lang.Class.newInstance (Class.java:429)
+   at java.lang.Class.newInstance (Class.java:421)
    at ProcessClassInst.main (ProcessClassInst.java:72)
 */
Index: test/regression/ProcessClassStop.java
===================================================================
RCS file: /cvs/kaffe/kaffe/test/regression/ProcessClassStop.java,v
retrieving revision 1.10
diff -u -p -r1.10 ProcessClassStop.java
--- test/regression/ProcessClassStop.java	5 Dec 2004 22:43:32 -0000	1.10
+++ test/regression/ProcessClassStop.java	19 Dec 2004 14:26:36 -0000
@@ -107,6 +107,6 @@ tan java.lang.NoClassDefFoundError: Base
 java.lang.NoClassDefFoundError: Base
    at java.lang.Class.getConstructor0 (Class.java)
    at java.lang.Class.getDeclaredConstructor (Class.java:174)
-   at java.lang.Class.newInstance (Class.java:429)
+   at java.lang.Class.newInstance (Class.java:421)
    at ProcessClassStop.main (ProcessClassStop.java:74)
 */
Index: test/regression/ProhibitedClass.java
===================================================================
RCS file: /cvs/kaffe/kaffe/test/regression/ProhibitedClass.java,v
retrieving revision 1.5
diff -u -p -r1.5 ProhibitedClass.java
--- test/regression/ProhibitedClass.java	14 Nov 2004 17:26:58 -0000	1.5
+++ test/regression/ProhibitedClass.java	19 Dec 2004 14:26:36 -0000
@@ -24,8 +24,8 @@ class ProhibitedClass
 
 /* Expected Output:
 java.lang.SecurityException: Prohibited package: java/lang/Object
-   at java.lang.ClassLoader.defineClass0 (ClassLoader.java)
-   at java.lang.ClassLoader.defineClass (ClassLoader.java:177)
-   at java.lang.ClassLoader.defineClass (ClassLoader.java:168)
+   at java.lang.VMClassLoader.defineClass (VMClassLoader.java)
+   at java.lang.ClassLoader.defineClass (ClassLoader.java:473)
+   at java.lang.ClassLoader.defineClass (ClassLoader.java:438)
    at ProhibitedClass.main (ProhibitedClass.java:21)
 */
